import abc
from copy import deepcopy
from itertools import product

from cobamp.core.models import ConstraintBasedModel


class ModelTransformer(object):
	__metaclass__ = abc.ABCMeta

	@staticmethod
	def transform(self, args, properties):

		# args must be:
		# - a dict with 'S', 'lb', 'ub' keys
		# - a ConstraintBasedModel
		if isinstance(args, dict):
			assert len(set(args.keys()) & {'S', 'lb', 'ub'}) == len(set(args.keys())), 'args must contain at least S' + \
																					   ', lb, and ub key-value pairs'

			S, lb, ub = [args[k] for k in ['S', 'lb', 'ub']]
			return self.transform_array(S, lb, ub, properties)

		elif isinstance(args, ConstraintBasedModel):
			S = args.get_stoichiometric_matrix()
			lb, ub = args.get_bounds_as_list()
			new_properties = deepcopy(properties)

			for k in ['block', 'keep']:
				if new_properties[k] != None:
					new_properties[k] = [args.decode_index(r, 'reaction') for r in properties[k]]

			Sn, lbn, ubn, mapping, metabs = self.transform_array(S, lb, ub, new_properties)

			reaction_names_new = [
				new_properties['reaction_id_sep'].join([args.reaction_names[i] for i in mapping.from_new(i)]) for i in
				range(len(lbn))]
			modeln = ConstraintBasedModel(
				S=Sn,
				thermodynamic_constraints=[list(k) for k in list(zip(lbn, ubn))],
				reaction_names=reaction_names_new,
				metabolite_names=[args.metabolite_names[k] for k in metabs]
			)

			return modeln, mapping, metabs

	@staticmethod
	@abc.abstractmethod
	def transform_array(S, lb, ub, properties):
		## TODO: implement

		# must return:
		# - new S matrix
		# - new lower/upper bounds
		# - mapping between rows/cols from both matrices
		# mapping = ReactionIndexMapping({}, {})
		# metabs = []
		# return S, lb, ub, mapping, metabs
		return


class ReactionIndexMapping(object):
	def __init__(self, otn, nto):
		self.otn = otn
		self.nto = nto

	def from_original(self, idx):
		return self.otn[idx]

	def from_new(self, idx):
		return self.nto[idx]

	def multiply(self, new_ids):
		return list(product(*[self.from_new(k) for k in set(new_ids)]))



import abc
import warnings
from collections import OrderedDict
from itertools import chain
from multiprocessing import cpu_count

import numpy as np
import optlang
from optlang.symbolics import Zero

CUSTOM_DEFAULT_SOLVER = None
SOLVER_ORDER = ['CPLEX', 'GUROBI', 'GLPK', 'MOSEK', 'SCIPY']


def get_default_solver():
	if CUSTOM_DEFAULT_SOLVER:
		return CUSTOM_DEFAULT_SOLVER
	else:
		for solver in SOLVER_ORDER:
			if optlang.list_available_solvers()[solver]:
				return solver


def get_solver_interfaces():
	s = {}
	for solver, status in optlang.list_available_solvers().items():
		if status:
			try:
				s[solver] = eval('optlang.' + solver.lower() + '_interface')
			except:
				print('Optlang reports',solver.lower(),'is available but the interface could not be imported.')
	return s


SOLVER_INTERFACES = get_solver_interfaces()
DEFAULT_SOLVER = get_default_solver()

VAR_CONTINUOUS, VAR_INTEGER, VAR_BINARY = ('continuous', 'integer', 'binary')
VAR_TYPES = (VAR_CONTINUOUS, VAR_INTEGER, VAR_BINARY)
VARIABLE, CONSTRAINT = 'var', 'const'
SENSE_MINIMIZE, SENSE_MAXIMIZE = ('min', 'max')


def fwd_irrev(lb, ub):
	"""
	Args:
	 lb:
	 ub:
	"""
	return ((lb >= 0) and (ub >= 0)).astype(int)


def bak_irrev(lb, ub):
	"""
	Args:
	 lb:
	 ub:
	"""
	return ((lb < 0) and (ub <= 0)).astype(int)


def fix_backwards_irreversible_reactions(S, lb, ub):
	"""
	Args:
	 S:
	 lb:
	 ub:
	"""
	S = np.array(S)
	lb, ub = np.array(lb), np.array(ub)
	fwd_irrev_index, bak_irrev_index = [[i for i in range(S.shape[1]) if fx(lb[i], ub[i])] for fx in
										[fwd_irrev, bak_irrev]]
	# irrev = np.union1d(fwd_irrev_index, bak_irrev_index)

	if len(bak_irrev_index) > 0:
		S[:, bak_irrev_index] = -S[:, bak_irrev_index]
		ub_temp = ub[bak_irrev_index]
		ub[bak_irrev_index] = -lb[bak_irrev_index]
		lb[bak_irrev_index] = -ub_temp

	return S, lb, ub, fwd_irrev_index, bak_irrev_index


def make_irreversible_model(S, lb, ub):
	# lb, ub = np.array(lb), np.array(ub)
	# # fwd_irrev = lambda lb, ub: (lb >= 0) and (ub >= 0)
	# # bak_irrev = lambda lb, ub: (lb < 0) and (ub <= 0)
	#
	# fwd_irrev_index, bak_irrev_index = [[i for i in range(S.shape[1]) if fx(lb[i], ub[i])] for fx in [fwd_irrev, bak_irrev]]
	"""
	Args:
	 S:
	 lb:
	 ub:
	"""
	S, lb, ub, fwd_irrev_index, bak_irrev_index = fix_backwards_irreversible_reactions(S, lb, ub)

	irrev = np.union1d(fwd_irrev_index, bak_irrev_index)

	# irrev = np.array([i for i in range(S.shape[1]) if not (lb[i] < 0 and ub[i] > 0)])
	rev = np.array([i for i in range(S.shape[1]) if i not in irrev])

	# if len(bak_irrev_index) > 0:
	# 	S[:, bak_irrev_index] = -S[:, bak_irrev_index]
	# 	ub_temp = ub[bak_irrev_index]
	# 	ub[bak_irrev_index] = -lb[bak_irrev_index]
	# 	lb[bak_irrev_index] = -ub_temp

	if len(rev) > 0:
		Sr = S[:, rev]
		S_new = np.hstack([S, -Sr])
	else:
		S_new = S.copy()
	offset = S.shape[1]
	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
	for i, rx in enumerate(rev):
		rx_mapping[rx].append(offset + i)
	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])

	nlb, nub = np.zeros(S_new.shape[1]), np.zeros(S_new.shape[1])
	for orig_rx, new_rx in rx_mapping.items():
		if isinstance(new_rx, tuple):
			nub[new_rx[0]] = abs(lb[orig_rx])
			nub[new_rx[1]] = ub[orig_rx]
		else:
			nlb[new_rx], nub[new_rx] = lb[orig_rx], ub[orig_rx]

	return S_new, nlb, nub, rx_mapping


class LinearSystem():
	"""An abstract class defining the template for subclasses implementing linear
	systems that can be used with optimizers such as CPLEX and passed onto other
	algorithms supplied with the package.

	Must instantiate the following variables:

	 S: System of linear equations represented as a n-by-m ndarray, preferrably
	 with dtype as float or int

	 __model: Linear model as an instance of the solver.
	"""
	__metaclass__ = abc.ABCMeta
	model = None

	def was_built(self):
		if self.model != None:
			return (len(self.model.variables) != 0) or (len(self.model.constraints) != 0)
		else:
			return False

	def select_solver(self, solver=None):
		"""
		Args:
		solver:
		"""
		if not solver:
			solver = get_default_solver()
			self.interface = SOLVER_INTERFACES[solver]
		else:
			if solver in SOLVER_INTERFACES:
				self.interface = SOLVER_INTERFACES[solver]

			else:
				raise Exception(solver, 'Solver not found. Choose from ', str(list(SOLVER_INTERFACES.keys())))
		self.solver = solver

	def get_configuration(self):
		return self.model.configuration

	def set_default_configuration(self):
		cfg = self.model.configuration

		try:
			cfg.tolerances.feasibility = 1e-9
			cfg.tolerances.optimality = 1e-9
			cfg.tolerances.integrality = 1e-9
			cfg.lpmethod = 'auto'
			cfg.presolve = True

		except:
			print('Could not set parameters with this solver')

	def get_constraint_bounds(self, constraints=None):
		"""
		Args:
		  constraints:
		"""
		return [(c.lb, c.ub) for c in (self.model.constraints if constraints == None else constraints)]

	def get_constraint_matrices(self, constraints=None, vars=None):
		"""
		Args:
		  constraints:
		  vars:
		"""
		return [self.get_system_matrix(constraints, vars)] + list(zip(*self.get_constraint_bounds(constraints)))

	def get_system_matrix(self, constraints=None, vars=None):
		"""
		Args:
		  constraints:
		  vars:
		"""
		var_list = self.model.variables if vars == None else vars
		cons_list = self.model.constraints if constraints == None else constraints
		var_dict = dict(list(zip(*list(zip(*enumerate(var_list)))[::-1])))
		cnst = cons_list

		def row_factory(mask, values):
			row = np.zeros([1, len(var_list)])
			# print(len(mask), len(values), row.shape)
			row[:, mask] = values
			return row

		def get_mask_value_list(constraint):
			if constraint.indicator_variable == None:
				return list(zip(
					*[(var_dict[var], value) for var, value in constraint.get_linear_coefficients(var_list).items()]))
			else:
				return [(0, 1), (0, 0)]

		A = np.vstack([row_factory(*get_mask_value_list(c)) for c in cnst])
		return A

	def set_number_of_threads(self, n_threads=0):
		"""Defines the amount of threads available for the solver to use. :param
		n_threads: Number of threads available to the solver. Set to 0 if default
		parameters are needed :return:

		Args:
		  n_threads:
		"""
		cpu_c = cpu_count()
		is_positive = n_threads >= 0
		actual_threads = n_threads

		if not is_positive:
			warnings.warn('Threads cannot be set to negative values. Using default values')
			actual_threads = 0

		if n_threads > cpu_c:
			warnings.warn('User-defined number of threads exceeds the amount of physical threads')

		if self.solver == 'CPLEX':
			self.model.problem.parameters.threads.set(actual_threads)
		elif self.solver == 'GUROBI':
			self.model.problem.Params.Threads = actual_threads
		else:
			warnings.warn('Could not set threads for ' + str(self.solver) + ' instance. Not yet implemented!')

	def set_working_memory_limit(self, workmem):
		"""Defines the amount of memory available for the solver to use. Use this at
		your own peril! :param n_threads: Memory in MegaBytes available to the solver
		:return:

		Args:
		  workmem:
		"""
		is_positive = workmem >= 0
		actual_mem = workmem

		if not is_positive:
			warnings.warn('Threads cannot be set to negative values. Using default values')

		else:
			if self.solver == 'CPLEX':
				self.model.problem.parameters.workmem.set(actual_mem)
			elif self.solver == 'GUROBI':
				self.model.problem.Params.NodefileStart = actual_mem / 1024
			else:
				warnings.warn('Could not set threads for ' + str(self.solver) + ' instance. Not yet implemented!')

	@abc.abstractmethod
	def build_problem(self):
		"""Builds a CPLEX model with the constraints specified in the constructor
		arguments. This method must be implemented by any <LinearSystem>. Refer to the
		:ref:`constructor <self.__init__>` -------
		"""
		pass

	def get_model(self):
		"""Returns the model instance. Must call <self.build_problem()> to return a
		CPLEX model.

		-------
		"""
		return self.model

	def get_stoich_matrix_shape(self):
		"""Returns a tuple with the shape (rows, columns) of the supplied
		stoichiometric matrix.

		-------
		"""
		return self.S.shape

	# def empty_constraint(self, b_lb, b_ub, dummy_var=Variable(name='dummy', lb=0, ub=0)):
	def empty_constraint(self, b_lb, b_ub, **kwargs):
		"""
		Args:
		  b_lb:
		  b_ub:
		  **kwargs:
		"""
		return self.interface.Constraint(Zero, lb=b_lb, ub=b_ub, **kwargs)

	def dummy_variable(self):
		return self.interface.Variable(name='dummy', lb=0, ub=0)

	def populate_model_from_matrix(self, S, var_types, lb, ub, b_lb, b_ub, var_names, only_nonzero=False,
								   indicator_rows=None):
		"""
		Args:
		  S: Two-dimensional np.ndarray instance
		  var_types: list or tuple with length equal to the amount of columns of S
			  containing variable types (e.g.
		  lb: list or tuple with length equal to the amount of columns of S
			  containing the lower bounds for all variables
		  ub: list or tuple with length equal to the amount of columns of S
			  containing the upper bounds for all variables
		  b_lb: list or tuple with length equal to the amount of rows of S
			  containing the lower bound for the right hand
		  b_ub: list or tuple with length equal to the amount of rows of S
			  containing the upper bound for the right hand
		  var_names: string identifiers for the variables
		  only_nonzero:
		  indicator_rows:
		"""
		self.add_variables_to_model(var_names, lb, ub, var_types)
		self.add_rows_to_model(S, b_lb, b_ub, only_nonzero, indicator_rows)

	def populate_constraints_from_matrix(self, S, constraints, vars, only_nonzero=False):
		"""
		Args:
		  S: Two-dimensional np.ndarray instance
		  constraints (side of all):
		  vars: list of variable instances
		  only_nonzero:
		"""
		if not only_nonzero:
			coef_list = [{vars[j]: S[i, j] for j in range(S.shape[1])} for i in range(S.shape[0])]
		else:
			coef_list = [{vars[j]: S[i, j] for j in np.nonzero(S[i, :])[0]} for i in range(S.shape[0])]

		for coefs, constraint in zip(coef_list, constraints):
			if constraint.indicator_variable is None:
				if len(coefs) > 0:
					constraint.set_linear_coefficients(coefs)

		self.model.update()

	def add_rows_to_model(self, S_new, b_lb, b_ub, only_nonzero=False, indicator_rows=None, vars=None, names=None):
		"""
		Args:
		  S_new:
		  b_lb:
		  b_ub:
		  only_nonzero:
		  indicator_rows:
		  vars:
		  names:
		"""
		if not vars:
			vars = self.model.variables
		# dummy = self.dummy_variable()
		if names != None:
			constraints = [self.empty_constraint(b_lb[i], b_ub[i], name=names[i]) for i in range(S_new.shape[0])]
		else:
			constraints = [self.empty_constraint(b_lb[i], b_ub[i]) for i in range(S_new.shape[0])]
		if indicator_rows:
			for row, var_idx, complement in indicator_rows:
				constraints[row] = self.interface.Constraint(
					sum(S_new[row, i] * vars[i] for i in S_new[row, :].nonzero()[0]), lb=b_lb[row], ub=b_ub[row],
					indicator_variable=vars[var_idx], active_when=complement)
		self.model.add(constraints, sloppy=True)

		self.model.update()

		self.populate_constraints_from_matrix(S_new, constraints, vars, only_nonzero)
		# self.model.update()
		# self.model.remove(dummy)
		self.model.update()
		return constraints

	def remove_from_model(self, index, what):
		"""
		Args:
		  index:
		  what:
		"""
		container = self.model.variables if what == VARIABLE else self.model.constraints if what == CONSTRAINT else None
		if type(index) not in (list, tuple):
			index = [index]
		for i in index:
			try:
				self.model.remove(container[i])
			except Exception as e:
				warnings.warn('Could not remove constraint with index '+str(i)+' with exception:'+str(e))

		if what == VARIABLE:
			self.S = np.delete(self.S, index, 1)
		elif what == CONSTRAINT:
			self.S = np.delete(self.S, index, 0)

		self.model.update()

	def add_columns_to_model(self, S_new, var_names, lb, ub, var_types, only_nonzero=False):
		"""
		Args:
		  S_new:
		  var_names:
		  lb:
		  ub:
		  var_types:
		"""
		vars = self.add_variables_to_model(var_names, lb, ub, var_types)

		self.populate_constraints_from_matrix(S_new, self.model.constraints, vars, only_nonzero=only_nonzero)

	def add_variables_to_model(self, var_names, lb, ub, var_types):

		"""
		Args:
		  var_names:
		  lb:
		  ub:
		  var_types:
		"""
		if isinstance(var_types, str):
			var_types = [var_types] * len(lb)

		vars = [self.interface.Variable(name=name, lb=lbv, ub=ubv, type=t) for name, lbv, ubv, t in
				zip(var_names, lb, ub, var_types)]
		self.model.add(vars)
		self.model.update()

		return vars

	def get_stuff(self, what, index):
		"""
		Args:
		  what:
		  index:
		"""
		container = self.model.variables if what == VARIABLE else self.model.constraints if what == CONSTRAINT else None
		if container == None:
			raise ValueError('`what` parameter requires a string (either `var` or `const`)')

		return [container[k] for k in index]

	def set_objective(self, coefficients, minimize, vars=None):
		"""
		Args:
		  coefficients:
		  minimize:
		  vars:
		"""
		if not vars:
			vars = self.model.variables
		lcoefs = {k: v for k, v in zip(vars, coefficients) if v != 0}
		if len(lcoefs) == 0:
			lcoefs = {k: 0 for k in vars}
		try:
			dummy = self.dummy_variable()
			new_obj = self.interface.Objective(dummy)
			self.model.objective = new_obj
			new_obj.set_linear_coefficients(lcoefs)
		except Exception as e:
			raise e
		finally:
			self.model.remove(dummy)
			self.model.objective.direction = SENSE_MINIMIZE if minimize else SENSE_MAXIMIZE
			self.model.update()

	def set_variable_bounds(self, vars, lb, ub):
		"""
		Args:
		  vars:
		  lb:
		  ub:
		"""
		for var, ulb, uub in zip(vars, lb, ub):
			var.set_bounds(ulb, uub)
		self.model.update()

	def set_constraint_bounds(self, constraints, lb, ub):
		"""
		Args:
		  constraints:
		  lb:
		  ub:
		"""
		for c, ulb, uub in zip(constraints, lb, ub):
			b = ulb, uub
			lb_is_greater = b[0] > c.ub if ((c.ub is not None) and (b[0] is not None)) else False
			ub_is_lower = b[1] < c.lb if ((c.lb is not None) and (b[1] is not None)) else False
			if lb_is_greater:
				c.ub = b[1]
				c.lb = b[0]
			elif ub_is_lower:
				c.lb = b[0]
				c.ub = b[1]
			else:
				c.lb, c.ub = b
		self.model.update()

	def set_variable_types(self, vars, types):
		"""
		Args:
		  vars:
		  types:
		"""
		if isinstance(types, str):
			for var in vars:
				var.type = types
		else:
			for var, typ in zip(vars, types):
				if typ in VAR_TYPES:
					var.type = typ
				else:
					warnings.warn('Invalid variable type: ' + typ)
		self.model.update()

	def write_to_lp(self, filename):
		"""
		Args:
		  filename:
		"""
		with open(filename, 'w') as f:
			f.write(self.model.to_lp())


class KShortestCompatibleLinearSystem(LinearSystem):
	"""Abstract class representing a linear system that can be passed as an
	argument for the KShortestAlgorithm class. Subclasses must instantiate the
	following variables:

	 __dvar_mapping: A dictionary mapping reaction indexes with variable names

	 __dvars: A list of variable names (str) or Tuple[str] if two linear system
	 variables represent a single flux. Should be kept as type `list` to
	 maintain order.

	 __c: str representing the variable to be used as constant for indicator
	 constraints
	"""
	dvar_mapping = None
	dvars = None
	c = None

	def get_dvar_mapping(self):
		"""Returns a dictionary mapping flux indices with variable(s) on the optimization problem
		-------
		"""
		return self.dvar_mapping

	def get_dvars(self):
		"""Returns a list of variables (str or Tuple[str]) with similar order to that of the fluxes passed to the system.
		-------
		"""
		return self.dvars

	def add_c_variable(self):
		self.c = self.add_variables_to_model(['C'], [1], [None], VAR_CONTINUOUS)[0]

	def get_c_variable(self):
		return self.model.variables['C']


class GenericLinearSystem(LinearSystem):
	"""Class representing a generic system of linear (in)equations Used as
	arguments for various algorithms implemented in the package.
	"""

	def __init__(self, S, var_types, lb, ub, b_lb, b_ub, var_names, solver=None):
		"""Constructor for GenericLinearSystem

		Parameters

		  model: Optlang model S: Two-dimensional np.ndarray instance var_types:
		  list or tuple with length equal to the amount of columns of S containing
		  variable types (e.g. VAR_CONTINUOUS) lb: list or tuple with length equal
		  to the amount of columns of S containing the lower bounds for all
		  variables ub: list or tuple with length equal to the amount of columns of
		  S containing the upper bounds for all variables b_lb: list or tuple with
		  length equal to the amount of rows of S containing the lower bound for the
		  right hand side of all constraints b_ub: list or tuple with length equal
		  to the amount of rows of S containing the upper bound for the right hand
		  side of all constraints var_names: string identifiers for the variables

		Args:
		  S:
		  var_types:
		  lb:
		  ub:
		  b_lb:
		  b_ub:
		  var_names:
		  solver:
		"""
		self.select_solver(solver)
		self.model = self.interface.Model()
		self.set_default_configuration()
		self.S, self.lb, self.ub, self.b_lb, self.b_ub, self.var_types = S, lb, ub, b_lb, b_ub, var_types

		self.names = var_names if var_names is not None else ['v' + str(i) for i in range(S.shape[1])]

	def build_problem(self):
		self.populate_model_from_matrix(self.S, self.var_types, self.lb, self.ub, self.b_lb, self.b_ub, self.names,
										True)


class SteadyStateLinearSystem(GenericLinearSystem):
	"""Class representing a steady-state biological system of metabolites and
	reactions without dynamic parameters Used as arguments for various algorithms
	implemented in the package.
	"""

	def __init__(self, S, lb, ub, var_names, solver=None):
		"""Constructor for SimpleLinearSystem

		Parameters

		----------

		  S: Stoichiometric matrix represented as a n-by-m ndarray, preferrably with
		  dtype as float or int

		  lb: ndarray or list containing the lower bounds for all n fluxes

		  ub: ndarray or list containing the lower bounds for all n fluxes

		  var_names: - optional - ndarray or list containing the names for each flux

		Args:
		  S:
		  lb:
		  ub:
		  var_names:
		  solver:
		"""
		m, n = S.shape
		self.lb, self.ub = lb, ub
		super().__init__(S, VAR_CONTINUOUS, lb, ub, [0] * m, [0] * m, var_names, solver=solver)


def prepare_irreversible_system(self, S, lb, ub, non_consumed, consumed, produced, non_produced, solver, force_bounds, add_c=True):
	"""
	Args:
	 self:
	 S:
	 lb:
	 ub:
	 non_consumed:
	 consumed:
	 produced:
	 solver:
	 force_bounds:
	"""
	self.select_solver(solver)

	# lb = [0 if i in irrev else -1 for i in range(S.shape[1])]
	# ub = [1] * S.shape[1]
	# if -1 in lb:
	S, lb, ub, fwd_irrev, bak_irrev = fix_backwards_irreversible_reactions(S, lb, ub)

	Si, lbi, ubi, rev_mapping = make_irreversible_model(S, lb, ub)
	fwd_names = ['V' + str(i) if not isinstance(v, list) else 'V' + str(i) + 'F' for i, v in rev_mapping.items()]
	bwd_names = ['V' + str(i) + 'R' for i, v in rev_mapping.items() if isinstance(v, (list, tuple))]

	# else:
	# 	Si, lbi, ubi = S, lb, ub
	# 	fwd_names = ['V' + str(i) for i in range(Si.shape[1])]
	# 	bwd_names = []
	# 	rev_mapping = {i:i for i in range(Si.shape[1])}

	lbi = [0] * len(lbi) + [1] if add_c else []
	ubi = [None] * len(ubi) + [None] if add_c else []

	for k, tup in force_bounds.items():
		if isinstance(rev_mapping[k], (list, tuple)):
			inds = rev_mapping[k]
			trev = (abs(ub[k]) if ub[k] < 0 else 0, abs(lb[k]) if lb[k] < 0 else 0)
			tfwd = (abs(lb[k]) if lb[k] > 0 else 0, abs(ub[k]) if ub[k] > 0 else 0)
			assert sum([pair[0] < pair[1] for pair in (trev, tfwd)]) == 2, 'force_bounds contains invalid values'
			for i, ntup in zip(inds, (tfwd, trev)):
				lbi[i], ubi[i] = ntup
		else:
			assert tup[0] < tup[1], 'force_bounds contains invalid values'
			lbi[k], ubi[k] = tup

	## TODO: remove this line - added for debugging
	# ubi = [10000 if ((k == None) or (k >= 10000)) else k for k in ubi]

	self.rev_mapping = rev_mapping
	self.__ivars = None
	self.__ss_override = [(nc, 'G', 0) for nc in non_consumed] + [(p, 'G', 1) for p in produced] + \
	                     [(c, 'L', -1) for c in consumed] + [(npd, 'L', 0) for npd in non_produced]

	Si = np.hstack([Si, np.zeros((Si.shape[0], 1))]) if add_c else Si
	b_lb, b_ub = [0] * Si.shape[0], [0] * Si.shape[0]
	## TODO: Maybe allow other values to be provided for constraint relaxation/tightening

	for ident, sign, v in self.__ss_override:
		if sign == 'G':
			b_lb[ident] = v if b_lb[ident] != None else None
			b_ub[ident] = None
		elif sign == 'L':
			b_lb[ident] = None
			b_ub[ident] = v if b_ub[ident] != None else None
	for i in range(len(b_lb)):
		if (b_lb[i] is None) and (b_ub[i] is None):
			b_lb[i], b_ub[i] = [-1e6, 1e6]

	return Si, lbi, ubi, b_lb, b_ub, fwd_names, bwd_names, rev_mapping


class IrreversibleLinearSystem(KShortestCompatibleLinearSystem, GenericLinearSystem):
	"""Class representing a steady-state biological system of metabolites and
	reactions without dynamic parameters. All irreversible reactions are split into
	their forward and backward components so every lower bound is 0. Used as
	arguments for various algorithms implemented in the package.
	"""

	def __init__(self, S, lb, ub, non_consumed=(), consumed=(), produced=(), non_produced=(), solver=None, force_bounds={}):
		"""
		Args:
		  S (Stoichiometric matrix represented as a n-by-m ndarray, preferrably with dtype as float or int):
		  lb:
		  ub:
		  non_consumed (An Iterable[int] or ndarray containing the indices of external metabolites not consumed in the):
		  consumed (An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be produced.):
		  produced (An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be consumed.):
		  solver (String specifying the LP solver to be used):
		  force_bounds (Dictionary mapping indexes with 2-tuples with lower and upper bounds):
		"""
		Si, lbi, ubi, b_lb, b_ub, fwd_names, bwd_names, rev_mapping = \
			prepare_irreversible_system(self, S, lb, ub, non_consumed, consumed, produced, non_produced, solver, force_bounds)

		super().__init__(Si, VAR_CONTINUOUS, lbi, ubi, b_lb, b_ub, fwd_names + bwd_names + ['C'], solver=solver)

		self.dvars = list(range(S.shape[1] + len(bwd_names)))
		self.dvar_mapping = dict(rev_mapping)


class IrreversibleLinearPatternSystem(IrreversibleLinearSystem):
	## TODO: Code + docstrings. Do not use this yet!
	def __init__(self, S, lb, ub, subset, **kwargs):
		"""
		Args:
		  S:
		  lb:
		  ub:
		  subset:
		  **kwargs:
		"""
		super().__init__(S, lb, ub, **kwargs)
		self.subset = subset
		fwds, revs = [], []
		dvm_new = {}
		for r in self.subset:
			if not isinstance(self.rev_mapping[r], int):
				fid, rid = self.dvar_mapping[r]
				dvm_new[r] = (len(fwds), len(subset) + len(revs))
				fwds.append(fid)
				revs.append(rid)
			else:
				dvm_new[r] = len(fwds)
				fwds.append(self.dvar_mapping[r])

		self.dvar_mapping = dvm_new
		self.dvars = fwds + revs

	def build_problem(self):
		super().build_problem()


class GenericDualLinearSystem(KShortestCompatibleLinearSystem, GenericLinearSystem):
	def __init__(self, S, K, T, b, solver=None):
		self.select_solver(solver)

		self.__ivars = None
		self.__c = "C"
		super().__init__(*self.generate_dual_problem(S, K, T, b), solver=solver)

		offset = S.shape[0]
		self.dvars = list(range(offset, offset + K.shape[0]))
		self.dvar_mapping = {i: i for i in range(K.shape[0])}

	def generate_dual_problem(self, S, K, T, b):
		m, n = S.shape
		veclens = [("u", m),("v", K.shape[0]),("w", T.shape[0])]

		var_prop_list = [[(pref + str(i), 0 if pref != "u" else None, None) for i in range(n)] for
						pref, n in
						veclens]


		Sdi = np.hstack([S.T, K.T, T.T, np.zeros([S.shape[1], 1])])
		Sd = np.vstack([Sdi, np.zeros([1, Sdi.shape[1]])])

		names, v_lb, v_ub = list(zip(*chain(*var_prop_list+[[(self.__c,1, None)]])))

		Sd[-1, [m + K.shape[0] + i for i in range(len(b))]] = b
		Sd[-1, -1] = 1

		b_ub = np.hstack([np.array([None] * Sdi.shape[0]), np.array([0])])
		b_lb = np.array([0] * (Sd.shape[0]), dtype=object)
		b_ub[-1] = 0
		b_lb[-1] = None


		return Sd, VAR_CONTINUOUS, v_lb, v_ub, b_lb, b_ub, names


class DualLinearSystem(KShortestCompatibleLinearSystem, GenericLinearSystem):
	"""Class representing a dual system based on a steady-state metabolic network
	whose elementary flux modes are minimal cut sets for use with the KShortest
	algorithms. Based on previous work by Ballerstein et al. and Von Kamp et al.

	[1] von Kamp, A., & Klamt, S. (2014). Enumeration of smallest intervention
	strategies in genome-scale metabolic networks. PLoS computational biology,
	10(1), e1003378. [2] Ballerstein, K., von Kamp, A., Klamt, S., & Haus, U. U.
	(2011). Minimal cut sets in a metabolic network are elementary modes in a dual
	network. Bioinformatics, 28(3), 381-387.
	"""

	def __init__(self, S, lb, ub, T, b, solver=None):
		"""Parameters

		----------

		  S: Stoichiometric matrix represented as a n-by-m ndarray, preferrably with
		  dtype as float or int irrev: An Iterable[int] or ndarray containing the
		  indices of irreversible reactions

		  T: Target matrix as an ndarray. Should have c-by-n dimensions (c -
		  #constraints; n - #fluxes)

		  b: Inhomogeneous bound values as a list or 1D ndarray of c size n.

		Args:
		  S:
		  lb:
		  ub:
		  T:
		  b:
		  solver:
		"""
		self.select_solver(solver)

		S, lb, ub, fwd_irrev, bak_irrev = fix_backwards_irreversible_reactions(S, lb, ub)
		irrev = np.union1d(fwd_irrev, bak_irrev).astype(int)

		self.__ivars = None
		self.S, self.irrev, self.T, self.b = S, irrev, T, b
		self.__c = "C"
		super().__init__(*self.generate_dual_problem(S, irrev, T, b), solver=solver)

		offset = S.shape[0]
		self.dvars = list(range(offset, offset + (S.shape[1] * 2)))

		self.dvar_mapping = {i: (i, S.shape[1] + i) for i in range(S.shape[1])}

	def generate_dual_problem(self, S, irrev, T, b):
		"""
		Args:
		  S:
		  irrev:
		  T:
		  b:
		"""
		m, n = S.shape
		Sxi, Sxr = S[:, irrev].T, np.delete(S, irrev, axis=1).T

		I = np.identity(n)
		Ii, Ir = I[irrev, :], np.delete(I, irrev, axis=0)

		veclens = [("u", m)] + [("vp", I.shape[0]), ("vn", I.shape[0])] + [("w", self.T.shape[0])]


		Ti, Tr = T[:, irrev].T, np.delete(T, irrev, axis=1).T

		var_prop_list = [[(pref + str(i), 0 if pref != "u" else None, None) for i in range(n)] for
						pref, n in
						veclens]


		Sdi = np.hstack([Sxi, Ii, -Ii, Ti, np.zeros([Sxi.shape[0], 1])])
		Sdr = np.hstack([Sxr, Ir, -Ir, Tr, np.zeros([Sxr.shape[0], 1])])
		Sd = np.vstack([Sdi, Sdr, np.zeros([1, Sdi.shape[1]])])

		names, v_lb, v_ub = list(zip(*chain(*var_prop_list)))

		names = list(names) + ['C']
		v_lb = list(v_lb) + [1]
		v_ub = list(v_ub) + [None]

		w_idx = [m + n + n + i for i in range(len(b))]
		Sd[-1, w_idx] = b
		Sd[-1, -1] = 1

		b_ub = np.hstack([np.array([None] * Sdi.shape[0]), np.array([0] * Sdr.shape[0]), np.array([0])])
		b_lb = np.array([0] * (Sd.shape[0]), dtype=object)
		b_ub[-1] = 0
		b_lb[-1] = None

		return Sd, VAR_CONTINUOUS, v_lb, v_ub, b_lb, b_ub, names


class BendersMasterSystem(GenericLinearSystem):
	def __init__(self, F, c, g, lb, ub, solver):
		"""
		Args:
		  F:
		  c:
		  g:
		  lb:
		  ub:
		  solver:
		"""
		var_names = ['x' + str(i) for i in range(F.shape[1])]
		super().__init__(S=F, var_types=VAR_INTEGER, lb=lb, ub=ub, b_lb=np.array([None] * F.shape[1]),
						 b_ub=g, var_names=var_names, solver=solver)
		self.c = c
		self.cuts = []

	def build_problem(self):
		super().build_problem()
		self.set_objective(self.c, True)

	def add_combinatorial_benders_cut(self, x_sol):
		"""
		Args:
		  x_sol:
		"""
		cut_coefs = x_sol.copy()
		pos_mask = cut_coefs > 1e-6
		cut_coefs[pos_mask] = -1
		cut_coefs[~pos_mask] = 1
		rhs = pos_mask.sum()
		self.cuts.extend(self.add_rows_to_model(cut_coefs.reshape(1, -1), [1 - rhs], [None]))

	def remove_cuts(self):
		self.remove_from_model(index=[c.name for c in self.cuts], what='const')


class BendersSlaveSystem(GenericLinearSystem):
	def __init__(self, A, M, D, b, e, lb_y, ub_y, solver=None):
		"""
		Args:
		  A:
		  M:
		  D:
		  b:
		  e:
		  lb_y:
		  ub_y:
		  solver:
		"""
		self.slack_vars = ['slack' + str(i) for i in range(A.shape[0])]
		self.y_var_names = ['y' + str(i) for i in range(A.shape[1])]
		var_names = self.y_var_names + self.slack_vars
		self.b_ub_fixed = np.array([None] * (A.shape[0] + D.shape[0]))
		self.lby, self.uby = np.array(lb_y), np.array(ub_y)
		Af = np.hstack([A, np.eye(A.shape[0])])
		Df = np.hstack([D, np.zeros([D.shape[0], A.shape[0]])])
		slack_bounds = np.zeros(A.shape[0])

		super().__init__(S=np.vstack([Af, Df]), var_types=VAR_CONTINUOUS,
						 lb=np.concatenate([self.lby, slack_bounds]),
						 ub=np.concatenate([self.uby, slack_bounds]),
						 b_lb=np.array([0] * (A.shape[0] + D.shape[0])), b_ub=self.b_ub_fixed,
						 var_names=var_names, solver=solver)

		self.e, self.M, self.b = e, M, b
		self.constraints = self.model.constraints
		self.y_var_mask = np.arange(0, A.shape[0])
		self.previous_changes = []

	def parametrize(self, x_sol):
		"""
		Args:
		  x_sol:
		"""
		if len(self.previous_changes) > 0:
			chgmap = zip(self.previous_changes, np.zeros(len(self.previous_changes)))
			self.model._set_variable_bounds_on_problem(chgmap, chgmap)

		rhs = np.concatenate([self.b - np.dot(self.M, x_sol.reshape(-1, 1)).ravel(), self.e.copy()], axis=0)
		nzi = np.nonzero(rhs)[0]
		self.previous_changes = [self.model.variables[self.slack_vars[k]] for k in nzi]
		values_to_change = rhs[nzi]
		nchgmap = zip(self.previous_changes, values_to_change)
		# self.set_constraint_bounds(self.constraints, lb=rhs, ub=self.b_ub_fixed)
		# lb_n, ub_n = np.concatenate([self.lby, -rhs]), np.concatenate([self.uby, -rhs])
		self.model._set_variable_bounds_on_problem(nchgmap, nchgmap)



def pretty_table_print(table, has_header=True, header_sep=2):
	table = list([list(t) for t in table])

	prntstr = ''
	col_sep_max = ((max([max([len(k) for k in it]) for it in table])//4)+1)
	col_sep = '\t'*col_sep_max

	# check for inconsistent dimensions

	if has_header:
		header = table[0]
		data = table[1:]
		prntstr += col_sep.join(header)+('\n'*header_sep)
	else:
		data = list(table)
	ndata = []
	for line in data:
		tokens = []
		for stri in line:
			separator = '\t'*(col_sep_max - (len(stri) // 4) + 1)
			tokens.extend([stri, separator])
		tokens.pop()
		ndata.append(tokens)

	prntstr += '\n'.join(''.join(k) for k in ndata)
	print(prntstr)


if __name__ == '__main__':
	pretty_table_print([['column1','column2'],['value19999','value2'],['item1item11999','item2'],['item3','']])


from itertools import product
from multiprocessing import cpu_count

from pathos.pools import _ProcessPool

from cobamp.core.optimization import BatchOptimizer

MP_THREADS = cpu_count()


def _batch_function(param_index):
    global _params, _function, _iterable
    return param_index, _function(_iterable[param_index], _params)


def _pool_initializer(params):
    global _iterable, _params, _function
    _params = params
    _iterable = params['iterable']
    _function = params['function']


def _batch_run(params, threads):
    jobs = len(params['iterable'])
    res_map = [None for _ in range(jobs)]
    true_threads = min((jobs // 2) + 1, threads)
    it_per_job = jobs // threads
    pool = _ProcessPool(
        processes=true_threads,
        initializer=_pool_initializer,
        initargs=([params])
    )
    for i, value in pool.imap_unordered(_batch_function, list(range(jobs)),
                                        chunksize=it_per_job):
        res_map[i] = value

    pool.close()
    pool.join()

    return res_map


def batch_run(function, sequence, paramargs=None, threads=MP_THREADS):
    params = {'function': function, 'iterable': sequence}
    if paramargs != None:
        params.update(paramargs)
    return _batch_run(params, min(len(sequence),threads))

def batch_optimize_cobamp_model(cobamp_model, bounds, objectives, combine_inputs=False, max_threads=MP_THREADS):

    def check_inputs(bounds, objectives):
        all_eq = len(bounds) == len(objectives)
        if combine_inputs or not all_eq:
            bounds, objectives = zip(*product(bounds, objectives))

        obj_coef, senses = zip(*objectives)

        bounds, obj_coef = [{cobamp_model.decode_index(k, 'reaction'): v for k, v in d.items()}
                            for d in (bounds, obj_coef)]

        return bounds, obj_coef, objectives

    cbounds, ccoefs, cobjs = check_inputs(bounds, objectives)
    bopt = BatchOptimizer(cobamp_model.model, threads=min(cbounds, max_threads))
    return bopt.batch_optimize(cbounds, ccoefs, cobjs)



from operator import add, sub, mul, truediv, pow


class Stack(list):

	def push(self, x):
		self.append(x)

	def top(self):
		return self[-1]


class Queue(list):

	def push(self, x):
		self.insert(0, x)

	def top(self):
		return self[-1]


def parse_infix_expression(op, is_operand_fx, is_operator_fx, precedence_fx):
	tokens = Queue(op[::-1])
	output = Queue()
	opstk = Stack()
	while tokens:
		token = tokens.pop()
		if is_operand_fx(token):
			output.push(token)
		elif is_operator_fx(token):
			while opstk and ((opstk.top() != '(') and ((precedence_fx(token) < precedence_fx(opstk.top())) or (
					left_operator_association(opstk.top()) and (precedence_fx(token) == precedence_fx(opstk.top()))))):
				output.push(opstk.pop())
			opstk.push(token)
		elif token == '(':
			opstk.push(token)
		elif token == ')':
			while opstk and (opstk.top() != '('):
				output.push(opstk.pop())
			if opstk.top() != '(':
				print('Mismatched parentheses found!')
			else:
				opstk.pop()
	while opstk:
		op_remaining = opstk.pop()
		if op_remaining in ('(', ')'):
			print('Mismatched parentheses found')
		output.push(op_remaining)
	return list(output)[::-1]


def evaluate_postfix_expression(op, eval_fx, type_conv=int):
	stk = Stack()
	for token in op:
		if is_operator_token(token):
			o1, o2 = type_conv(stk.pop()), type_conv(stk.pop())
			result = eval_fx(token, o1, o2)
			stk.push(result)
		elif is_number_token(token):
			stk.push(token)
	return stk.pop()


def tokenize_infix_expression(inf_exp_str):
	return list(filter(lambda x: x != '', inf_exp_str.replace('(', ' ( ').replace(')', ' ) ').split(' ')))


# def tokenize_boolean_expression(inf_exp_str, default_value='1'):
# 	return [tok if tok in ('0', '1', 'and', 'not', 'or', ')', '(') else default_value for tok in
# 			tokenize_infix_expression(inf_exp_str)]

def tokenize_boolean_expression(inf_exp_str):
	return tokenize_infix_expression(inf_exp_str)


def is_number_token(token):
	return token.replace('.', '', 1).replace('-', '', 1).isnumeric()


def is_operator_token(token):
	return token in ['**', '/', '*', '+', '-']


def is_string_token(token):
	return isinstance(token, str) and not is_boolean_operator(token) and (('(' not in token) and (')' not in token))


def op_prec(op):
	precedence = {
		'**': 4,
		'/': 3,
		'*': 3,
		'+': 2,
		'-': 1}
	return precedence[op]


def left_operator_association(op):
	return False if op == '**' else True


operators = {
	'+': add,
	'-': sub,
	'*': mul,
	'/': truediv,
	'**': lambda a, b: pow(b, a)
}


def eval_math_operator(operator, o1, o2):
	return operators[operator](o1, o2)


def is_boolean_value(token):
	if token in ('1', '0'):
		return True
	elif token.isnumeric():
		print('Illegal token', token, 'found.')
		return False
	else:
		return False


def is_boolean_operator(token):
	return token.upper() in ['AND', 'OR', 'NOT']


def boolean_precedence(token):
	pdict = {'not': 3,
			 'and': 2,
			 'or': 1}
	return pdict[token]


truth_table = {
	'and': {
		(0, 0): 0,
		(1, 0): 0,
		(0, 1): 0,
		(1, 1): 1
	},
	'or': {
		(0, 0): 0,
		(1, 0): 1,
		(0, 1): 1,
		(1, 1): 1
	},
	'not': {
		0: 1,
		1: 0
	}
}


def eval_boolean_operator(operator, o1, o2):
	return truth_table[operator][(o1, o2)]


if __name__ == '__main__':

	# def test_algebraic_expression():
	# 	op = tokenize_infix_expression('((15 / (7 - (1 + 1))) * 3) - (2 + (1 + 1)) ')
	# 	psfix = parse_infix_expression(op, is_number_token, is_operator_token, op_prec)
	# 	res = evaluate_postfix_expression(psfix, eval_math_operator)
	# 	print(res)


	# def test_boolean_expression():
	from urllib.request import urlretrieve
	from cobra.io import read_sbml_model
	from random import random

	BOOL_OPS = ('and', 'not', 'or', ')', '(')
	path, content = urlretrieve('http://bigg.ucsd.edu/static/models/RECON1.xml')
	model = read_sbml_model(path)

	# [(i,r.gene_name_reaction_rule) for i,r in enumerate(model.reactions)]
	ogpr = model.reactions[616].gene_name_reaction_rule
	gpr_tok = tokenize_infix_expression(ogpr)
	psfix = parse_infix_expression(gpr_tok, lambda x: x not in BOOL_OPS, is_boolean_operator, boolean_precedence)

	# gene_activation = {k: 1 for k in [g.id for g in model.genes]}
	# for test in range(20):
	# 	gpr = ogpr
	# 	for gene in gene_activation:
	# 		dec = random() > 0.6
	# 		gpr = gpr.replace(gene, '1' if dec else '0')
	# 	op = [tok if tok in ('0', '1', 'and', 'not', 'or', ')', '(') else '1' for tok in
	# 		  tokenize_infix_expression(gpr)]
	# 	psfix = parse_infix_expression(op, is_boolean_value, is_boolean_operator, boolean_precedence)
	# 	print(gpr, ''.join(psfix), evaluate_postfix_expression(psfix, eval_boolean_operator), sep=',')

	# test_boolean_expression()


from cobamp.algorithms.kshortest import *
from cobamp.core.linear_systems import IrreversibleLinearSystem, DualLinearSystem, IrreversibleLinearPatternSystem, \
	GenericDualLinearSystem

from itertools import product

from cobamp.wrappers import get_model_reader


class KShortestEnumeratorWrapper(object):
	__metaclass__ = abc.ABCMeta
	"""
	An abstract class for methods involving the K-shortest EFM enumeration algorithms
	"""
	ALGORITHM_TYPE_ITERATIVE = 'kse_iterative'
	ALGORITHM_TYPE_POPULATE = 'kse_populate'

	__alg_to_prop_name = {
		ALGORITHM_TYPE_ITERATIVE: K_SHORTEST_OPROPERTY_MAXSOLUTIONS,
		ALGORITHM_TYPE_POPULATE: K_SHORTEST_OPROPERTY_MAXSIZE
	}

	__alg_to_alg_name = {
		ALGORITHM_TYPE_ITERATIVE: K_SHORTEST_METHOD_ITERATE,
		ALGORITHM_TYPE_POPULATE: K_SHORTEST_METHOD_POPULATE
	}

	def __init__(self, model, algorithm_type=ALGORITHM_TYPE_POPULATE, stop_criteria=1, forced_solutions=None,
				 excluded_solutions=None, solver='CPLEX', force_bounds={}, n_threads=0, workmem=None, big_m=False,
				 max_populate_sols_override=None, time_limit=None, big_m_value=None, cut_function=None, extra_args=None,
				 pre_enum_function=None):
		"""

		Parameters

		----------

			model: A Model instance from the external framework to use. Must be registered in the dict stored as
			external_wrappers.model_readers along with its reader.

			algorithm_type: ALGORITHM_TYPE_ITERATIVE or ALGORITHM_TYPE_POPULATE constants stored as class attributes.
				ALGORITHM_TYPE_ITERATIVE is a slower method (regarding EFMs per unit of time) that enumerates EFMs one
				at a time.
				ALGORITHM_TYPE_POPULATE enumerates EFMs one size at a time. This is the preferred method as it's
				generally faster.

			stop_criteria: An integer that defines the stopping point for EFM enumeration. Either refers to the maximum
			number of EFMs or the maximum size they can reach before the enumeration stops.

			forced_solutions: A list of KShortestSolution or lists of reaction indexes that must show up in the
			enumeration process. (experimental feature)

			excluded_solutions: A list of KShortestSolution or lists of reaction indexes that cannot show up in the
			enumeration process. (experimental feature)

			force_bounds: A dict mapping reaction indexes (int for now) with tuples containing lower and upper bounds
			An experimental feature meant to force certain phenotypes on EFP/EFMs

			n_threads: An integer value defining the amount of threads available to the solver

			workmem: An integer value defining the amount of memory in MegaBytes available to the solver
		"""

		self.__model = get_model_reader(model)
		self.model_reader = self.__model

		self.__algo_properties = KShortestProperties()
		self.__algo_properties[K_SHORTEST_MPROPERTY_METHOD] = self.__alg_to_alg_name[algorithm_type]
		self.__algo_properties[K_SHORTEST_MPROPERTY_TYPE_EFP] = self.is_efp
		self.__algo_properties[K_SHORTEST_OPROPERTY_N_THREADS] = n_threads
		self.__algo_properties[K_SHORTEST_OPROPERTY_WORKMEMORY] = workmem
		self.__algo_properties[K_SHORTEST_OPROPERTY_TIMELIMIT] = 0 if time_limit == None else time_limit
		self.__algo_properties[K_SHORTEST_OPROPERTY_BIG_M_CONSTRAINTS] = big_m
		self.__algo_properties[self.__alg_to_prop_name[algorithm_type]] = stop_criteria
		if big_m_value != None:
			self.__algo_properties[K_SHORTEST_OPROPERTY_BIG_M_VALUE] = big_m_value
		if (max_populate_sols_override != None) and algorithm_type == self.ALGORITHM_TYPE_POPULATE:
			self.__algo_properties[K_SHORTEST_OPROPERTY_MAXSOLUTIONS] = max_populate_sols_override


		def preprocess_cuts(cut_list):
			if cut_list is not None:
				new_cuts = []
				for cut in cut_list:
					new_cut = cut
					if not isinstance(cut, KShortestSolution):
						new_cut = [self.model_reader.reaction_id_to_index(k) if isinstance(k, str) else k for k in cut]
					new_cuts.append(new_cut)
				return new_cuts
			else:
				return []

		self.__forced_solutions = preprocess_cuts(forced_solutions)
		self.__excluded_solutions = preprocess_cuts(excluded_solutions)

		self.force_bounds = {self.model_reader.r_ids.index(k): v for k, v in force_bounds.items()}
		self.solver = solver
		self.pre_enum_function = pre_enum_function
		self.cut_function = cut_function
		self.extra_args = extra_args
		self.__setup_algorithm()
		self.enumerated_sols = []

	def __setup_algorithm(self):
		"""
		Creates the algorithms instance

		Returns:

		"""
		self.algo = KShortestEFMAlgorithm(self.__algo_properties, False)

	def __get_forced_solutions(self):
		"""
		Returns: A list of KShortestSolution or lists of reaction indexes

		"""
		return self.__forced_solutions

	def __get_excluded_solutions(self):
		"""
		Returns: A list of KShortestSolution or lists of reaction indexes

		"""
		return self.__excluded_solutions

	@abc.abstractmethod
	def get_linear_system(self):
		"""

		Returns a KShortestCompatibleLinearSystem instance build from the model
		-------

		"""
		return

	def get_enumerator(self):
		"""
		Returns an iterator that yields a single EFM or a list of multiple EFMs of the same size. Call next(iterator) to
		obtain the next set of EFMs.
		"""
		enumerator = self.algo.get_enumerator(
			linear_system=self.get_linear_system(),
			forced_sets=self.__get_forced_solutions(),
			excluded_sets=self.__get_excluded_solutions())

		if self.pre_enum_function is not None:
			self.pre_enum_function(self.algo.ksh, self.extra_args)
		for solarg in enumerator:
			self.enumerated_sols.append(solarg)
			yield self.decode_solution(solarg)
			if self.cut_function is not None:
				cut_arg = solarg if isinstance(solarg, (tuple, list)) else [solarg]
				self.cut_function(cut_arg, self.algo.ksh, self.extra_args)

	def decode_k_shortest_solution(self, sol):
		## TODO: Make MAX_PRECISION a parameter for linear systems or the KShortestAlgorithm
		return {self.model_reader.r_ids[k]: sol.attribute_value(sol.SIGNED_VALUE_MAP)[k]
				for k in sol.get_active_indicator_varids()}

	def decode_solution(self, solarg):
		if isinstance(solarg, (list,tuple)):
			return [self.decode_k_shortest_solution(sol) for sol in solarg]
		else:
			return self.decode_k_shortest_solution(solarg)

class KShortestEFMEnumeratorWrapper(KShortestEnumeratorWrapper):
	"""
	Extension of the abstract class KShortestEnumeratorWrapper that takes a metabolic model as input and yields
	elementary flux modes.
	"""

	def __init__(self, model, non_consumed, consumed, produced, non_produced, subset=None, **kwargs):
		"""

		Parameters

		----------

			model: A Model instance from the external framework to use. Must be registered in the dict stored as
			external_wrappers.model_readers along with its reader.

			non_consumed: An Iterable[int] or ndarray containing the indices of external metabolites not consumed in the
			model.

			consumed: An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be produced.

			produced: An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be consumed.


		"""
		self.is_efp = False
		super().__init__(model, **kwargs)
		self.__consumed, self.__non_consumed, self.__produced, self.__non_produced, self.__subset = \
			consumed, non_consumed, produced, non_produced, subset

	def get_linear_system(self):
		to_convert = [self.__consumed, self.__non_consumed, self.__produced, self.__non_produced]
		conv_cn, conv_nc, conv_pr, conv_np = [[self.model_reader.metabolite_id_to_index(k) for k in lst] for lst in to_convert]
		lb, ub = [array(k) for k in self.model_reader.get_model_bounds(as_dict=False, separate_list=True)]
		if self.__subset == None:

			return IrreversibleLinearSystem(
				S=self.model_reader.S,
				lb=lb, ub=ub,
				consumed=conv_cn,
				non_consumed=conv_nc,
				produced=conv_pr,
				non_produced=conv_np,
				solver=self.solver,
				force_bounds=self.force_bounds
			)
		else:
			return IrreversibleLinearPatternSystem(
				S=self.model_reader.S,
				lb=lb, ub=ub,
				consumed=conv_cn,
				non_consumed=conv_nc,
				produced=conv_pr,
				non_produced=conv_np,
				subset=[self.model_reader.reaction_id_to_index(s) for s in self.__subset],
				solver=self.solver,
				force_bounds=self.force_bounds
			)

class KShortestEFPEnumeratorWrapper(KShortestEnumeratorWrapper):
	"""
	Extension of the abstract class KShortestEnumeratorWrapper that takes a metabolic model as input and yields
	elementary flux patterns.
	"""

	def __init__(self, model, subset, non_consumed=[], consumed=[], produced=[], non_produced=[],**kwargs):
		self.is_efp = True
		super().__init__(model, **kwargs)
		self.__subset = subset
		self.__consumed, self.__non_consumed, self.__produced, self.__non_produced = consumed, non_consumed, produced, non_produced

	def get_linear_system(self):
		## TODO:  change irrev to lb/ub structure
		to_convert = [self.__consumed, self.__non_consumed, self.__produced, self.__non_produced]
		lb, ub = [array(k) for k in self.model_reader.get_model_bounds(as_dict=False, separate_list=True)]
		conv_cn, conv_nc, conv_pr, conv_np = [[self.model_reader.metabolite_id_to_index(k) for k in lst] for lst in to_convert]
		conv_subsets = [self.model_reader.reaction_id_to_index(s) for s in self.__subset]
		return IrreversibleLinearPatternSystem(
			S=self.model_reader.S,
			lb=lb, ub=ub,
			subset=conv_subsets,
			consumed=conv_cn,
			non_consumed=conv_nc,
			produced=conv_pr,
			non_produced=conv_np,
			solver=self.solver,
			force_bounds=self.force_bounds
		)




class KShortestMCSEnumeratorWrapper(KShortestEnumeratorWrapper):
	"""
	Extension of the abstract class KShortestEnumeratorWrapper that takes a metabolic model as input and yields
	minimal cut sets.

	"""

	def __init__(self, model, target_flux_space_dict, target_yield_space_dict, **kwargs):
		self.is_efp = False
		super().__init__(model, **kwargs)
		self.__ip_constraints = list(chain(*AbstractConstraint.convert_tuple_intervention_problem(
			target_flux_space_dict, target_yield_space_dict, self.model_reader)))

	def materialize_intv_problem(self):
		return InterventionProblem(self.model_reader.S).generate_target_matrix(self.__ip_constraints)

	def get_linear_system(self):
		lb, ub = [array(k) for k in self.model_reader.get_model_bounds(separate_list=True, as_dict=False)]
		T, b = self.materialize_intv_problem()
		return DualLinearSystem(self.model_reader.S, lb, ub, T, b, solver=self.solver)


class KShortestGenericMCSEnumeratorWrapper(KShortestEnumeratorWrapper):

	def __init__(self, model, target_flux_space_dict, target_yield_space_dict, dual_matrix, dual_var_mapper, **kwargs):
		self.is_efp = False
		super().__init__(model, **kwargs)
		self.__ip_constraints = list(chain(*AbstractConstraint.convert_tuple_intervention_problem(
			target_flux_space_dict, target_yield_space_dict, self.model_reader)))

		self.dual_matrix, self.dual_var_mapper = dual_matrix, {v:k for k,v in dual_var_mapper.items()}

	def decode_k_shortest_solution(self, sol):
		## TODO: Make MAX_PRECISION a parameter for linear systems or the KShortestAlgorithm
		mapper = self.dual_var_mapper if self.dual_var_mapper is not None else self.model_reader.r_ids
		return {mapper[k]: sol.attribute_value(sol.SIGNED_VALUE_MAP)[k] for k in sol.get_active_indicator_varids()}

	def get_linear_system(self):
		T, b = InterventionProblem(self.model_reader.S).generate_target_matrix(self.__ip_constraints)
		return GenericDualLinearSystem(self.model_reader.S, self.dual_matrix, T, b, solver=self.solver)


class KShortestGeneticMCSEnumeratorWrapper(KShortestGenericMCSEnumeratorWrapper):
	@staticmethod
	def gene_cut_function(solx, ksh, extra_args):
		alternative_gene_identity = extra_args['F']

		for sol in solx:
			act_vars = sol.get_active_indicator_varids()
			if len(act_vars) > 1:
				dependencies = list(chain(*product(filter(lambda x: len(x) > 0,
														  [list(alternative_gene_identity[av]) for av in act_vars]))))
			else:
				dependencies = [[k] for k in [alternative_gene_identity[av] for av in act_vars][0]]
			if len(dependencies) > 0:
				ksh.exclude_solutions(dependencies)

	@staticmethod
	def set_gene_weights(ksh, extra_args):
		ksh.set_objective_expression(extra_args['gene_weights'])

	def __init__(self, model, target_flux_space_dict, target_yield_space_dict, G, gene_map, F, gene_weights, **kwargs):
		super().__init__(model, target_flux_space_dict, target_yield_space_dict, G, gene_map,
						 cut_function=self.gene_cut_function, pre_enum_function=self.set_gene_weights,
						 extra_args={'gene_map': gene_map, 'F': F, 'gene_weights':gene_weights}, **kwargs)


"""
Inspired by Metatool's code
"""
from itertools import chain

from numpy import sqrt, triu, logical_not, nonzero, mean, zeros, argmin, isin, sign, delete, unique, where, \
	dot, eye, setdiff1d, ndarray, array
from numpy.linalg import norm
from cobamp.core.transformer import ModelTransformer, ReactionIndexMapping
from cobamp.utilities.property_management import PropertyDictionary
from cobamp.nullspace.nullspace import compute_nullspace, nullspace_blocked_reactions

EPSILON = 2 ** -52
PRECISION = 1e-10


def subset_reduction(S, irrev, to_remove=[], to_keep_single=[]):
	"""
	Reduces a stoichiometric matrix using nullspace analysis by identifying linearly dependent (enzyme) subsets.
	These reactions are then compressed.

	Parameters
	----------
		S: Stoichiometric matrix as an ndarray.

		irrev: A boolean array with size equal to the number of columns in the S matrix.

		to_remove: A list of indices specifying columns of the S matrix to remove before the compression (usually blocked
		reactions)

		to_keep_single: A list of indices specifying columns of the S matrix not to compress.

	Returns	rd, sub, irrev_reduced, rdind, irrv_subsets, kept_reactions, kernel, correlation_matrix

		rd : compressed stoichiometric matrix -> numpy.array

		sub : subset matrix, n-subsets by n-reactions -> numpy.array

		irrev_reduced : subset reversibilities -> numpy.array of type bool

		rdind : metabolite indices -> numpy.array of type int

		irrv_subsets : same as sub, but list if empty

		kept_reactions : indexes for reactions used in the network compression

		kernel : numpy.array with the right nullspace of S

		correlation_matrix : numpy.array with reaction correlation matrix
	-------

	"""
	m, n = S.shape

	keep_single = array([False] * n)
	keep_single[to_keep_single] = True

	kept_reactions = array([True] * n)
	kept_reactions[to_remove] = False
	kept_reactions = where(kept_reactions)[0]

	ktol = EPSILON * sum(kept_reactions)

	kernel = compute_nullspace(S[:, kept_reactions], ktol, False)
	kernel_blocked = nullspace_blocked_reactions(kernel, ktol)

	if kernel_blocked.shape[0] > 0:
		kept_reactions = kept_reactions[setdiff1d(kept_reactions, kernel_blocked)]
		kernel = compute_nullspace(S[:, kept_reactions], ktol, False)

	correlation_matrix = subset_candidates(kernel)
	S_scm = S[:, kept_reactions]
	irrev_scm = irrev[kept_reactions]
	scm_kp_ids = where([keep_single[kept_reactions]])[1]

	sub, irrev_reduced, irrv_subsets = subset_correlation_matrix(S_scm, kernel, irrev_scm, correlation_matrix,
																 scm_kp_ids)
	if len(kept_reactions) < n:
		temp = zeros([sub.shape[0], n])
		temp[:, kept_reactions] = sub
		sub = temp
		if len(irrv_subsets) > 0:
			temp = zeros([len(irrv_subsets), n])
			temp[:, kept_reactions] = irrv_subsets
			irrv_subsets = temp
	rd, rdind, irrev_reduced_final, sub = reduce(S, sub, irrev_reduced)

	return rd, sub, irrev_reduced_final, rdind, irrv_subsets, kept_reactions, kernel, correlation_matrix


def subset_candidates(kernel, tol=None):
	"""
	Computes a matrix of subset candidates from the nullspace of the S matrix

	Parameters

	----------

		kernel: Nullspace of the S matrix

		tol: Tolerance to 0.

	Returns a 2D triangular ndarray
	-------

	"""
	tol = kernel.shape[0] * EPSILON if tol is None else tol
	cr = dot(kernel, kernel.T)
	for i in range(kernel.shape[0]):
		for j in range(i + 1, kernel.shape[0]):
			cr[i, j] = cr[i, j] / sqrt(cr[i, i] * cr[j, j])
		cr[i, i] = 1
	cr = triu(cr)
	cr[abs(abs(cr) - 1) >= tol] = 0
	return sign(cr)


def subset_correlation_matrix(S, kernel, irrev, cr, keepSingle=None):
	"""

	Parameters
	----------
		S: Stoichiometric matrix as ndarray

		kernel: The nullspace of S

		irrev: List of booleans representing irreversible reactions (when True)

		cr: The subset candidate matrix, computed using <subset_candidates>

		keepSingle: List of reaction indices that will not be compressed.

	Returns sub, irrev_sub, irrev_violating_subsets

		sub : subset matrix, n-subsets by n-reactions -> numpy.array

		irrev_sub : subset reversibilities -> numpy.array of type bool

		irrev_violating_subsets : same as sub, but list if empty. Contains subsets discarded due to irreversibility faults
	-------

	"""
	m, n = S.shape
	in_subset = array([False] * n)
	irrev_sub = array([False] * cr.shape[0])
	sub = zeros([cr.shape[0], n])
	irrev_violating_subsets = []
	sub_count = 0
	if (keepSingle is not None) and (len(keepSingle) > 0):
		# keepSingle = array([])
		irrev_violating_subsets = []
		sub[:len(keepSingle), keepSingle] = eye(len(keepSingle))
		irrev_sub[:len(keepSingle)] = irrev[keepSingle]
		in_subset[keepSingle] = True
		sub_count = len(keepSingle)
	for i in range(cr.shape[0] - 1, -1, -1):
		reactions = where(cr[:, i] != 0)[0]
		in_subset_indexes = where(in_subset)[0]
		in_subset_reactions = isin(reactions, in_subset_indexes)
		reactions = reactions[logical_not(in_subset_reactions)]
		if len(reactions) > 0:
			in_subset[reactions] = True
			irrev_sub[sub_count] = (irrev[reactions]).any()

			if len(reactions) == 1:
				sub[sub_count, reactions] = 1
			else:
				lengths = norm(kernel[reactions, :], axis=1)
				min_ind = argmin(abs(lengths - mean(lengths)))
				lengths /= lengths[min_ind]
				sub[sub_count, reactions] = lengths * cr[reactions, i]
			sub_count += 1

	sub = sub[:sub_count, :]
	irrev_sub = irrev_sub[:sub_count]

	ind = unique(where(sub[:, irrev] < 0)[0])
	if len(ind) > 0:
		sub[ind, :] = -sub[ind, :]
		ind = unique(where(sub[:, irrev] < 0)[0])
		if len(ind) > 0:
			irrev_violating_subsets = sub[ind, :]
			sub = delete(sub, ind, 0)
			irrv_to_keep = delete(array(range(len(irrev_sub))), ind, 0)
			irrev_sub = irrev_sub[irrv_to_keep]

	return sub, irrev_sub, irrev_violating_subsets


def reduce(S, sub, irrev_reduced=None):
	"""
	Reduces a stoichiometric matrix according to the subset information present in the sub matrix and irrev_reduced.

	Parameters

	----------

		S: Stoichiometric matrix

		sub: Subset matrix as computed by <subset_correlation_matrix>

		irrev_reduced: Irreversibility vector regarding the subsets.

	Returns reduced, reduced_indexes, irrev_reduced

	-------

	"""

	reduced = dot(S, sub.T)
	reduced[abs(reduced) < PRECISION] = 0
	reduced_indexes = unique(nonzero(reduced)[0])
	reduced = reduced[reduced_indexes, :]

	rdm, rdn = reduced.shape
	if rdn == 0 or rdm == 0:
		reduced = zeros(1, rdn)

	if irrev_reduced is not None:
		ind = unique(nonzero(reduced)[1])
		reduced = reduced[:, ind]
		irrev_reduced = irrev_reduced[ind]
		sub = sub[ind, :]
	else:
		irrev_reduced = []

	return reduced, reduced_indexes, irrev_reduced, sub


class SubsetReducerProperties(PropertyDictionary):
	def __init__(self, keep=None, block=None, absolute_bounds=False, reaction_id_sep='_+_'):
		def is_list(x):
			return isinstance(x, (tuple, list, ndarray))

		new_optional = {
			'keep': lambda x: is_list(x) or type(x) == None,
			'block': lambda x: is_list(x) or type(x) == None,
			'absolute_bounds': bool,
			'reaction_id_sep': str
		}

		super().__init__(optional_properties=new_optional)
		for name, value in zip(['keep', 'block', 'absolute_bounds', 'reaction_id_sep'],
							   [keep, block, absolute_bounds, reaction_id_sep]):
			self.add_if_not_none(name, value)


class SubsetReducer(ModelTransformer):
	TO_KEEP_SINGLE = 'SUBSET_REDUCER-TO_KEEP_SINGLE'
	TO_BLOCK = 'SUBSET_REDUCER-TO_BLOCK'
	ABSOLUTE_BOUNDS = 'SUBSET_REDUCER-ABSOLUTE_BOUNDS'

	@staticmethod
	def reduce(S, lb, ub, keep=(), block=(), absolute_bounds=False):
		lb, ub = list(map(array, [lb, ub]))
		to_keep, to_block = [], []
		irrev = (lb >= 0) | ((ub <= 0) & (lb <= 0))

		if block:
			to_block = array(block)

		if keep:
			to_keep = array(keep)

		rd, sub, irrev_reduced, rdind, irrv_subsets, kept_reactions, K, _ = subset_reduction(
			S, irrev, to_keep_single=to_keep, to_remove=to_block)

		mapping = SubsetReducer.get_transform_maps(sub)

		nlb = [0 if irrev_reduced[k] else None for k in range(rd.shape[1])]
		nub = [None] * rd.shape[1]

		if absolute_bounds:
			nlb = [0 if irrev_reduced[k] else -1000 for k in range(rd.shape[1])]
			nub = [float('inf')] * rd.shape[1]
			alb, aub = list(zip(*[[fx([x[k] for k in mapping.from_new(i)]) for x, fx in zip([lb, ub], [max, min])]
								  for i in range(rd.shape[1])]))

			for func, pair in zip([max, min], [[nlb, alb], [nub, aub]]):
				new, absolute = pair
				for i, v in enumerate(absolute):
					new[i] = func(new[i], absolute[i])

		return rd, nlb, nub, mapping, rdind

	@staticmethod
	def transform_array(S, lb, ub, properties):
		k, b, a = (properties[k] for k in ['keep', 'block', 'absolute_bounds'])

		Sn, lbn, ubn, mapping, metabs = SubsetReducer.reduce(S, lb, ub, k, b, a)

		return Sn, lbn, ubn, mapping, metabs

	@staticmethod
	def get_transform_maps(sub):
		new_to_orig = {i: list(nonzero(sub[i, :])[0]) for i in range(sub.shape[0])}
		orig_to_new = dict(chain(*[[(i, k) for i in v] for k, v in new_to_orig.items()]))

		return ReactionIndexMapping(orig_to_new, new_to_orig)



import warnings
import numpy as np

from cobamp.wrappers.core import AbstractObjectReader
from cobra.exceptions import Infeasible, Unbounded, UndefinedSolution

def cobra_simulate(cobra_model, cobra_func, bound_change, objective_coefficient, minimize, result_func, func_args=None):
    if func_args is None:
        func_args = {}
    with cobra_model as context_model:
        if bound_change is not None:
            for k, v in bound_change.items(): context_model.reactions.get_by_id(k).bounds = v

        if objective_coefficient is not None:
            context_model.objective = {context_model.reactions.get_by_id(k):v for k,v in objective_coefficient.items()}

        if minimize is not None:
            context_model.objective.direction = 'min' if minimize else 'max'
        try:
            sol = cobra_func(model=context_model, **func_args)
        except Exception as e:
            if isinstance(e, (Infeasible, Unbounded, UndefinedSolution)):
                sol = None
            else:
                raise e
        finally:
            return result_func(sol)

def cobra_simulation_result_function(sol):
    if sol is not None:
        return (sol.status == 'optimal', sol.objective_value, sol.fluxes.to_dict())
    else:
        return (False, None, None)

def cobra_fba(model, **func_args):
    return model.optimize()

class COBRAModelObjectReader(AbstractObjectReader):

    def __read_model(self, path, format, **kwargs):
        from cobra.io import read_sbml_model, load_matlab_model, load_json_model
        parse_functions = {
            'xml': read_sbml_model,
            'mat': load_matlab_model,
            'json': load_json_model,
            'sbml': read_sbml_model
        }
        if format == None:
            nformat = path.split('.')[-1]
        else:
            nformat = format
        if nformat in parse_functions.keys():
            return parse_functions[nformat](path, **kwargs)
        else:
            raise ValueError('Format '+str(nformat)+' is invalid or not yet available through the cobrapy readers. '+
                             'Choose one of the following: '+','.join(parse_functions.keys()))

    def __init__(self, model, gpr_gene_parse_function=str, format=None, **kwargs):
        if isinstance(model, str):
            warnings.warn('Reading model with cobrapy from the provided path...')
            model = self.__read_model(model, format, **kwargs)
        super().__init__(model, gpr_gene_parse_function=gpr_gene_parse_function, **kwargs)

    def get_stoichiometric_matrix(self):
        S = np.zeros((len(self.m_ids), len(self.r_ids)))
        for i, r_id in enumerate(self.r_ids):
            for metab, coef in self.model.reactions.get_by_id(r_id).metabolites.items():
                S[self.m_ids.index(metab.id), i] = coef

        return S

    def get_model_bounds(self, as_dict=False, separate_list=False):
        bounds = [r.bounds for r in self.rx_instances]
        if as_dict:
            return dict(zip(self.r_ids, bounds))
        else:
            if separate_list:
                return [list(bounds) for bounds in list(zip(*tuple(bounds)))]
            else:
                return tuple(bounds)

    def get_irreversibilities(self, as_index):
        irrev = [not r.reversibility for r in self.rx_instances]
        if as_index:
            irrev = list(np.where(irrev)[0])
        return irrev

    def get_rx_instances(self):
        return [self.model.reactions.get_by_id(rx) for rx in self.r_ids]

    def get_reaction_and_metabolite_ids(self):
        return tuple([[x.id for x in lst] for lst in (self.model.reactions, self.model.metabolites)])

    def get_model_genes(self):
        return set([g.id for g in self.model.genes])

    def get_model_gpr_strings(self, apply_fx=None):
        return [apply_fx(r.gene_reaction_rule) if apply_fx is not None
                else r.gene_reaction_rule for r in self.model.reactions]



import random
import string
from collections import OrderedDict
from time import time

import pandas as pd
from numpy import nan, array, abs, zeros, max
from pathos.multiprocessing import cpu_count
from pathos.pools import _ProcessPool

from cobamp.core.linear_systems import LinearSystem

MP_THREADS = cpu_count()


def random_string_generator(N):
	"""

	Parameters

	----------

		N : an integer

	Returns a random string of uppercase character and digits of length N
	-------

	"""
	return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(N))


class Solution(object):
	"""
	Class representing a solution to a given linear optimization problem. Includes an internal dictionary for additional
	information to be included.
	"""

	def __init__(self, value_map, status, **kwargs):
		"""

		Parameters
		----------

			value_map: A dictionary mapping variable indexes with their values as determined by the solver

			status: An object (preferrably str or int) containing the solution status

			kwargs: Any additional information to be included in the attribute_dict variable that can be accessed by
			the <self.attribute_value> function.
		"""
		self.__var_names = None
		self.__value_map = value_map
		self.__status = status
		self.__attribute_dict = {k: v for k, v in kwargs.items() if k != 'names'}
		if 'names' in kwargs:
			self.__var_names = kwargs['names']

		self.__obj_value = kwargs['objective_value'] if 'objective_value' in kwargs else nan

	def __getitem__(self, item):
		if hasattr(item, '__iter__') and not isinstance(item, str):
			return {k: self.__value_map[k] for k in item}
		elif isinstance(item, str):
			return self.__value_map[item]
		else:
			raise TypeError('\'item\' is not a sequence or string.')

	def to_series(self):
		if self.__var_names != None:
			return pd.Series({self.__var_names[i]: self.__value_map[i] for i in range(len(self.__var_names))})
		else:
			return pd.Series(self.var_values())

	def set_attribute(self, key, value):
		"""
		Sets the value of a given <key> as <value>.

		Parameters

		----------

			key - A string

			value - Any object to be associated with the supplied key
		-------

		"""
		self.__attribute_dict[key] = value

	def var_values(self):
		"""

		Returns a dict mapping reaction indices with the variable values.
		-------

		"""
		return self.__value_map

	def status(self):
		"""

		Returns the status of this solution, as supplied by the solver.
		-------

		"""
		return self.__status

	def attribute_value(self, attribute_name):
		"""

		Parameters

		----------

			attribute_name: A dictionary key (preferrably str)

		Returns the value associated with the supplied key
		-------

		"""
		return self.__attribute_dict[attribute_name]

	def attribute_names(self):
		"""

		Returns all keys present in the attribute dictionary.

		-------

		"""
		return self.__attribute_dict.keys()

	def objective_value(self):
		"""

		Returns the objective value for this solution

		"""
		return self.__obj_value

	def x(self):
		'''

		Returns a ndarray with the solution values in order (from the variables)

		'''

		return array(list(self.__value_map.values()))

	def __repr__(self):
		return '<'+self.status().capitalize()+' Solution - objective: '+\
			   str(self.objective_value())+'; at '+hex(id(self))+'>'


class LinearSystemOptimizer(object):
	"""
	Class with methods to solve a <LinearSystem> as a linear optimization problem.
	"""

	def __init__(self, linear_system, hard_fail=False, build=True):
		"""

		Parameters

		----------

			linear_system: A <LinearSystem> instance.
			hard_fail: A boolean flag indicating whether an Exception is raised when the optimization fails
		"""
		self.linear_system = linear_system
		if build:
			linear_system.build_problem()
		self.solver = linear_system.solver
		self.model = linear_system.get_model()
		self.hard_fail = hard_fail

	def optimize(self):
		"""
		Internal function to instantiate the solver and return a solution to the optimization problem

		Parameters

		----------

			objective: A List[Tuple[coef,name]], where coef is an objective coefficient and name is the name of the variable
			to be optimized.

			minimize: A boolean that, when True, defines the problem as a minimization

		Returns a <Solution> instance
		-------
		"""
		names = self.model._get_variables_names()

		value_map = OrderedDict([(v, nan) for v in names])
		status = None
		ov = nan

		# self.model.configuration.tolerances.feasibility = 1e-9 # TODO this is for a test, to delete later
		# self.model.configuration.tolerances.optimality = 1e-6 # TODO this is for a test, to delete later

		# tINIT test parameters
		# self.model.problem.Params.MIPGap = 1e-9
		# self.model.configuration.tolerances.feasibility = 1e-8
		# self.model.configuration.tolerances.optimality = 1e-8
		# self.model.configuration.verbosity = 3

		try:
			self.model.optimize()
			values = self.model._get_primal_values()
			value_map = OrderedDict([(k, v) for k, v in zip(names, values)])
			status = self.model.status
			ov = self.model.objective.value

		except Exception as e:
			frozen_exception = e

		if status or not self.hard_fail:
			return Solution(value_map, self.model.status, objective_value=ov)
		else:
			raise frozen_exception

	def populate(self, limit=None):
		intf_dict = {
			'CPLEX': self.__populate_cplex,
			'GUROBI': self.__populate_gurobi
		}
		if self.solver in ['CPLEX', 'GUROBI']:
			return intf_dict[self.solver](limit)
		else:
			raise ValueError('The provided solver does not have an implemented populate function. Choose from' +
							 ''.join(list(intf_dict.keys())))

	def __populate_cplex(self, limit):
		instance = self.model.problem

		if not limit:
			instance.parameters.mip.pool.capacity = instance.parameters.mip.pool.capacity.max()
		else:
			instance.parameters.mip.pool.capacity = limit
		vnames = instance.variables.get_names()
		mnames = self.model._get_variables_names()
		solutions = []
		try:
			instance.populate_solution_pool()
			pool_intf = instance.solution.pool
			nsols = pool_intf.get_num()
			for s in range(nsols):
				vmap = {k: v for k, v in zip(vnames, pool_intf.get_values(s))}
				ord_vmap = OrderedDict([(k, vmap[k]) for k in mnames])
				sol = Solution(ord_vmap, 'optimal', objective_value=pool_intf.get_objective_value(s))
				# TODO: get status dict from optlang and use it accordingly
				solutions.append(sol)
		except Exception as e:
			print(e)

		return solutions

	def __populate_gurobi(self, limit):

		instance = self.model.problem

		solutions = []
		instance.params.PoolSolutions = limit
		instance.params.SolutionNumber = 0
		try:
			instance.optimize()
			mnames = self.model._get_variables_names()
			if instance.SolCount > 0:
				for n in range(instance.SolCount):
					instance.params.SolutionNumber = n
					ord_vmap = OrderedDict([(k, instance.getVarByName(k).Xn) for k in mnames])
					sol = Solution(ord_vmap, 'optimal', objective_value=instance.PoolObjVal)
					solutions.append(sol)
		except Exception as e:
			print(e)
		finally:
			instance.params.SolutionNumber = 0

		return solutions


class BendersSlaveOptimizer(LinearSystemOptimizer):
	def __init__(self, slave_system, hard_fail, build):
		super().__init__(slave_system, hard_fail, build)

	def optimize(self):
		normal_sol = super().optimize()
		vmap = OrderedDict(zip(self.linear_system.y_var_names, normal_sol.x()[self.linear_system.y_var_mask]))
		return Solution(value_map=vmap, status=normal_sol.status())


class BendersDecompositionOptimizer(object):
	def __init__(self, master_system, slave_system, hard_fail=False, build=True):
		self.master, self.slave = master_system, slave_system
		t0 = time()
		self.opt_master, self.opt_slave = [opti(system, hard_fail, build) for opti, system in
										   zip([LinearSystemOptimizer, BendersSlaveOptimizer],
											   [self.master, self.slave])]
		t1 = time()
		print(t1 - t0, 'spent building the linear problems.')
		self.__set_model_parameters()
		self.master.remove_cuts()
		self.previous_cut = 0

	def _benders_iteration(self, master_sol):
		if master_sol.status() != 'infeasible':
			t2 = time()
			self.slave.parametrize(master_sol.x())
			t3 = time()
			print(t3 - t2, 'seconds spent applying parameters to the slave problem.')

			t4 = time()
			slave_sol = self.opt_slave.optimize()
			t5 = time()
			print(t5 - t4, 'seconds spent optimizing the slave problem.')
			# print('\tSlave has solution with status',slave_sol.status())
			print('Cutting at length = ', master_sol.x().astype(bool).sum())
			if slave_sol.status() == 'optimal':
				self.master.add_combinatorial_benders_cut(master_sol.x())
				return slave_sol
			else:
				# print('\tAdding cut with dimension =',master_sol.x().sum())
				self.master.add_combinatorial_benders_cut(master_sol.x())
				return None
		else:
			raise Exception('Master problem is infeasible. No further solutions.')

	def optimize(self, max_iterations=10000, slave_pool=20):
		i = 0
		r = None
		sol_stack = []
		while (i < max_iterations) and r == None:
			if len(sol_stack) > 1:
				master_sol = sol_stack.pop(0)
				r = self._benders_iteration(master_sol)
			else:
				t0 = time()
				sol_stack.extend(self.opt_master.populate(slave_pool))
				t1 = time()
				print(t1 - t0, 'spent populating the solution stack. Current length =', len(sol_stack))
			i += 1
		return r

	def __set_model_parameters(self):
		"""
		Sets the optlang Model instance's parameters with appropriate values for k-shortest enumeration
		"""
		parset_func = {'CPLEX': self.__set_model_parameters_cplex,
					   'GUROBI': self.__set_model_parameters_gurobi}

		if self.master.solver in parset_func.keys():
			parset_func[self.master.solver]()

	def __set_model_parameters_cplex(self):

		"""
		Internal method to set model parameters for the CPLEX solver. This is based on the original MATLAB code by Von
		Kamp et al.

		-------
		"""
		instance = self.master.model.problem

		instance.parameters.mip.tolerances.integrality.set(1e-9)
		# instance.parameters.mip.tolerances.mipgap.set(1e-2)
		# instance.parameters.mip.strategy.probe.set(3)
		instance.parameters.clocktype.set(1)
		instance.parameters.advance.set(1)
		instance.parameters.mip.strategy.fpheur.set(1)
		instance.parameters.emphasis.mip.set(2)
		instance.parameters.mip.pool.intensity.set(4)
		instance.parameters.mip.pool.absgap.set(0)
		instance.parameters.mip.pool.replace.set(2)

	def __set_model_parameters_gurobi(self):
		"""
		Internal method to set model parameters for the GUROBI solver. This is based on the original MATLAB code by Von
		Kamp et al.

		"""
		instance = self.master.model.problem

		instance.params.PoolGap = 0
		instance.params.MIPFocus = 2
		instance.params.MIPAbsGap = 0
		instance.params.PoolSearchMode = 2


def optimization_pool(lsystem, bound_change_list, objective_coef_list, objective_sense_list, threads=MP_THREADS):
	res_map = [None for _ in range(len(bound_change_list))]
	true_threads = min((len(bound_change_list) // 2) + 1, threads)
	it_per_job = len(bound_change_list) // threads
	pool = _ProcessPool(
		processes=true_threads,
		initializer=_pool_initializer,
		initargs=(lsystem, bound_change_list, objective_coef_list, objective_sense_list)
	)
	for i, value in pool.imap_unordered(_optimize_function, list(range(len(bound_change_list))),
										chunksize=it_per_job):
		res_map[i] = value

	pool.close()
	pool.join()
	return res_map


def _pool_initializer(linear_system: LinearSystem, bound_change_list, objective_coef_list, objective_sense_list):
	global _linear_system, _optimizer, _bound_change_list, _vars, _orig_lb, _orig_ub, _objective_coef_list, _objective_sense_list

	_linear_system, _bound_change_list, _objective_coef_list, _objective_sense_list = \
		linear_system, bound_change_list, objective_coef_list, objective_sense_list
	_vars = _linear_system.get_model().variables
	_orig_lb, _orig_ub = list(zip(*[(var.lb, var.ub) for var in _vars]))
	_optimizer = LinearSystemOptimizer(_linear_system, build=not _linear_system.was_built())


def _optimize_function(change_index):
	global _linear_system, _optimizer, _bound_change_list, _vars, _orig_lb, _orig_ub, _objective_coef_list, _objective_sense_list
	var_ids, bounds = list(zip(*list(_bound_change_list[change_index].items())))
	chg_vars = _linear_system.get_stuff(index=var_ids, what='var')
	lb, ub = list(zip(*bounds))
	olb, oub = [[l[k] for k in var_ids] for l in [_orig_lb, _orig_ub]]

	obj_var_ids, obj_coefs = list(zip(*list(_objective_coef_list[change_index].items())))

	obj_sense = _objective_sense_list[change_index]
	obj_vars = _linear_system.get_stuff(index=obj_var_ids, what='var')

	_linear_system.set_objective(coefficients=obj_coefs, minimize=obj_sense, vars=obj_vars)
	_linear_system.set_variable_bounds(chg_vars, lb, ub)
	sol = _optimizer.optimize()
	_linear_system.set_variable_bounds(chg_vars, olb, oub)

	return change_index, sol


class BatchOptimizer(object):
	def __init__(self, linear_system: LinearSystem, threads=MP_THREADS):
		self.__linear_system = linear_system
		self.__threads = threads

	def batch_optimize(self, bounds, objective_coefs, objective_senses):
		assert len(bounds) == len(objective_coefs) == len(objective_senses)
		return optimization_pool(self.__linear_system, bounds, objective_coefs, objective_senses,
								 threads=self.__threads)


class KShortestSolution(Solution):
	"""
	A Solution subclass that also contains attributes suitable for elementary flux modes such as non-cancellation sums
	of split reactions and reaction activity.
	"""
	SIGNED_INDICATOR_SUM = 'signed_indicator_map'
	SIGNED_VALUE_MAP = 'signed_value_map'

	def __init__(self, value_map, status, indicator_map, dvar_mapping, dvars, **kwargs):
		"""

		Parameters

		----------

			value_map: A dictionary mapping variable names with values

			status: See <Solution>

			indicator_map: A dictionary mapping indicators with

			dvar_mapping: A mapping between reaction indices and solver variables (Tuple[str] or str)

			kwargs: See <Solution>

		"""
		signed_value_map = {
			i: value_map[dvars[varlist[0]]] - value_map[dvars[varlist[1]]] if isinstance(varlist, (tuple, list)) else
			value_map[dvars[varlist]] for
			i, varlist in dvar_mapping.items()}
		signed_indicator_map = {
			i: value_map[indicator_map[dvars[varlist[0]]]] - value_map[indicator_map[dvars[varlist[1]]]] if isinstance(
				varlist,
				(tuple, list)) else
			value_map[indicator_map[dvars[varlist]]] for
			i, varlist in dvar_mapping.items()}
		super().__init__(value_map, status, **kwargs)
		self.set_attribute(self.SIGNED_VALUE_MAP, signed_value_map)
		self.set_attribute(self.SIGNED_INDICATOR_SUM, signed_indicator_map)

	def get_active_indicator_varids(self):
		"""

		Returns the indices of active indicator variables (maps with variables on the original stoichiometric matrix)

		-------

		"""
		return [k for k, v in self.attribute_value(self.SIGNED_INDICATOR_SUM).items() if abs(v) > 1e-9]



def identify_conflicts(modelnorm):
	"""
	Identify conflicting constraints in a constraint-based model instantiated with CPLEX
	:param modelnorm: A ConstraintBasedModel instance
	:return: A dict with reaction/metabolite names involved in an irreducible inconsistent subset and their bounds
			(if applicable)
	"""
	try:
		cpxi = modelnorm.model.model.problem
		cpxi.conflict.refine(cpxi.conflict.all_constraints())
		conf = {k[1]: modelnorm.get_reaction_bounds(k[1]) if 'bound' in k[0] else () for k, v in
				dict(zip([(cpxi.conflict.constraint_type[k[1][0][0]], modelnorm.reaction_names[k[1][0][1]]
				if cpxi.conflict.constraint_type[k[1][0][0]] != 'linear' else modelnorm.metabolite_names[k[1][0][1]])
						  for k in cpxi.conflict.get_groups()],
						 [cpxi.conflict.group_status[i] for i in cpxi.conflict.get()])).items()
				if v != 'excluded'}

		# for m in {k: v for k, v in conf.items() if v == ()}.keys():
		# 	matx = modelnorm.get_stoichiometric_matrix(rows=[m])[
		# 		modelnorm.get_stoichiometric_matrix(rows=[m]).nonzero()]
		# 	nzi = modelnorm.get_stoichiometric_matrix(rows=[m]).nonzero()[0]
		# 	invrx = [[c * v for v in vk] for vk, c in
		# 			 zip([modelnorm.get_reaction_bounds(k) for k in nzi], matx.tolist())]
		# 	final_metab_bound = [sum(k) for k in zip(*[(min(x), max(x)) for x in invrx])]
		# 	#print('\t', m, final_metab_bound, '>=', 1)

		#conf_rxs = {i: j for i, j in conf.items() if j != ()}
		return conf
	except:
		print('\tNo conflict')
	finally:
		modelnorm.revert_to_original_bounds()



from cobamp.wrappers.core import AbstractObjectReader
import numpy as np

def cobamp_simulate(cobamp_model, cobamp_func, bound_change, objective_coefficient, minimize, result_func=None,
                    func_args=None):

    if func_args is None:
        func_args = {}
    if result_func is None:
        result_func = lambda x: x

    with cobamp_model as context_model:
        if bound_change is not None:
            for k, v in bound_change.items(): context_model.set_reaction_bounds(k, lb=v[0], ub=v[1])

        if None not in [objective_coefficient,minimize]:
            context_model.set_objective(objective_coefficient, minimize)

        sol = cobamp_func(model=context_model, **func_args)
        return result_func(sol)

def cobamp_simulation_result_function(sol):
    return sol.status() == 'optimal', sol.objective_value(), sol.to_series().to_dict()

def cobamp_fba(model, **func_args):
    return model.optimize()

class CobampModelObjectReader(AbstractObjectReader):

    def get_stoichiometric_matrix(self):
        return self.model.get_stoichiometric_matrix()

    def get_model_bounds(self, as_dict, separate_list=False):
        if as_dict:
            return dict(zip(self.r_ids, self.model.bounds))
        else:
            if separate_list:
                return [bounds for bounds in list(zip(*tuple(self.model.bounds)))]
            else:
                return tuple(self.model.bounds)

    def get_irreversibilities(self, as_index):
        irrev = [not self.model.is_reversible_reaction(r) for r in self.r_ids]
        if as_index:
            irrev = list(np.where(irrev)[0])
        return irrev

    def get_reaction_and_metabolite_ids(self):
        return self.model.reaction_names, self.model.metabolite_names

    def get_rx_instances(self):
        return None

    def get_model_gpr_strings(self):
        return [self.model.gpr[i] for i in range(len(self.r_ids))]


from cobamp.wrappers.core import AbstractObjectReader
import numpy as np


class FramedModelObjectReader(AbstractObjectReader):

    def get_stoichiometric_matrix(self):
        return np.array(self.model.stoichiometric_matrix())

    def get_model_bounds(self, as_dict=False, separate_list=False):
        bounds = [(r.lb, r.ub) for r in self.rx_instances]
        if as_dict:
            return dict(zip(self.r_ids, bounds))
        else:
            if separate_list:
                return [bounds for bounds in list(zip(*tuple(bounds)))]
            else:
                return tuple(bounds)

    def get_irreversibilities(self, as_index):
        irrev = [not r.reversible for r in self.rx_instances]
        if as_index:
            irrev = list(np.where(irrev)[0])
        return irrev

    def get_reaction_and_metabolite_ids(self):
        return tuple(self.model.reactions.keys()), tuple(self.model.metabolites.keys())

    def get_rx_instances(self):
        return [self.model.reactions[rx] for rx in self.r_ids]

    def get_model_gpr_strings(self):
        return [rx.gene_reaction_rule for rx in self.get_rx_instances()]



import warnings
from .cobra import COBRAModelObjectReader
from .framed import FramedModelObjectReader
from .cobamp import CobampModelObjectReader
from .core import MatFormatReader, AbstractObjectReader

from ..wrappers import get_model_reader, model_readers

warnings.warn(
    '''\nThe wrappers.external_wrappers module will be deprecated in a future release in favour of the wrappers module. 
    Available ModelObjectReader classes can still be loaded using cobamp.wrappers.<class>. An appropriate model 
    reader can also be created using the get_model_reader function on cobamp.wrappers''')



import warnings
from collections import OrderedDict, Counter
from copy import deepcopy

from numpy import ndarray, array, delete, zeros, vstack, hstack, nonzero, append, int_, int8, int16, \
	int32, int64, where, isin

from cobamp.core.linear_systems import SteadyStateLinearSystem, VAR_CONTINUOUS, make_irreversible_model
from cobamp.utilities.context import CommandHistory
from cobamp.utilities.printing import pretty_table_print
from cobamp.core.optimization import LinearSystemOptimizer
from cobamp.core.cb_analysis import FluxVariabilityAnalysis
from cobamp.gpr.core import GPRContainer

INT_TYPES = (int, int_, int8, int16, int32, int64)
LARGE_NUMBER = 10e6 - 1
SMALL_NUMBER = 1e-6
BACKPREFIX = 'flux_backwards'

def to_list_if_single(x, n):
	if isinstance(x, (list, tuple)):
		return x
	else:
		return [x]*n

def make_irreversible_model_raven(S, lb, ub, inverse_reverse_reactions=False):
	irrev = array([i for i in range(S.shape[1]) if not (lb[i] < 0)])
	rev = array([i for i in range(S.shape[1]) if i not in irrev])
	Sr = S[:, rev]
	offset = S.shape[1]
	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
	for i, rx in enumerate(rev):
		rx_mapping[rx].append(offset + i)
	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])

	S_new = hstack([S, -Sr])
	nlb, nub = zeros(S_new.shape[1]), zeros(S_new.shape[1])
	for orig_rx, new_rx in rx_mapping.items():
		if isinstance(new_rx, tuple):
			nlb[new_rx[0]] = lb[orig_rx] if lb[orig_rx] >= 0 else 0  # not necessary since they're already 0
			nub[new_rx[0]] = ub[orig_rx] if ub[orig_rx] >= 0 else 0  # not necessary since they're already 0
			nlb[new_rx[1]] = ub[orig_rx] * -1 if (ub[orig_rx] * -1) >= 0 else 0
			nub[new_rx[1]] = lb[orig_rx] * -1 if (lb[orig_rx] * -1) >= 0 else 0

		else:
			nlb[new_rx] = lb[orig_rx]
			nub[new_rx] = ub[orig_rx]

	return S_new, nlb, nub, rx_mapping


# def make_irreversible_model(S, lb, ub):
# 	irrev = array([i for i in range(S.shape[1]) if not (lb[i] < 0 and ub[i] > 0)])
# 	rev = array([i for i in range(S.shape[1]) if i not in irrev])
# 	Sr = S[:, rev]
# 	offset = S.shape[1]
# 	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
# 	for i, rx in enumerate(rev):
# 		rx_mapping[rx].append(offset + i)
# 	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])
#
# 	S_new = hstack([S, -Sr])
# 	nlb, nub = zeros(S_new.shape[1]), zeros(S_new.shape[1])
# 	for orig_rx, new_rx in rx_mapping.items():
# 		if isinstance(new_rx, tuple):
# 			nub[new_rx[0]] = abs(lb[orig_rx])
# 			nub[new_rx[1]] = ub[orig_rx]
# 		else:
# 			nlb[new_rx], nub[new_rx] = lb[orig_rx], ub[orig_rx]
#
# 	return S_new, nlb, nub, rx_mapping


# def test_irreversible_conversions():
# 	S = zeros([3,3])
# 	lb, ub = ([0,1000],[-1000, 1000],[-1000, 0])


class ConstraintBasedModel(object):
	def __init__(self, S, thermodynamic_constraints, reaction_names=None, metabolite_names=None, optimizer=True,
				 solver=None, gprs=None):


		def __validate_args():

			# stoichiometric matrix
			if isinstance(S, ndarray):
				m, n = S.shape
			elif isinstance(S, tuple) or isinstance(S, list):
				m, n = len(S), len(S[0])
			else:
				raise TypeError('S matrix is not an ndarray or list')

			thermodynamic_types = set(map(type, thermodynamic_constraints))
			allowed_types = {list, tuple, bool}

			assert len(thermodynamic_types & allowed_types) == len(thermodynamic_types), 'Invalid bound type found.'

			reactions_ok = reaction_names is None or len(reaction_names) == n and isinstance(reaction_names, list)
			metabolites_ok = metabolite_names is None or len(metabolite_names) == m and isinstance(metabolite_names,
																								   list)

			assert reactions_ok, 'Reaction name dimensions do not match the supplied stoichiometrix matrix'
			assert metabolites_ok, 'Metabolite name dimensions do not match the supplied stoichiometrix matrix'
			return m, n

		self.solver = solver
		m, n = __validate_args()
		self.__S = array(S)

		self.bounds = self.__interpret_bounds(thermodynamic_constraints)
		self.original_bounds = deepcopy(self.bounds)
		self.reaction_names, self.metabolite_names = deepcopy(reaction_names), deepcopy(metabolite_names)

		self.__update_decoder_map()

		if gprs is not None:
			self.gpr = gprs
		else:
			self.gpr = ['']*len(self.reaction_names)
		self.c = None
		self.model = None

		if optimizer:
			self.initialize_optimizer()

		self.context_managers = []

	def has_context(self):
		return len(self.context_managers) > 0

	def __enter__(self):
		self.context_managers.append(CommandHistory())
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		ctx_man = self.context_managers.pop()
		ctx_man.execute_all(False)

	def __getstate__(self):
		return self.__dict__

	def __update_decoder_map(self):
		self.reaction_decoder_map = self.metabolite_decoder_map = None

		if self.reaction_names:
			self.reaction_decoder_map = dict(zip(self.reaction_names, list(range(self.__S.shape[1]))))
		if self.metabolite_names:
			self.metabolite_decoder_map = dict(zip(self.metabolite_names, list(range(self.__S.shape[0]))))

		self.map_labels = {'reaction': self.reaction_decoder_map,
						   'metabolite': self.metabolite_decoder_map}

	def __interpret_bounds(self, thermodynamic_constraints):
		def interpret_bound(bound_obj):
			i, bound = bound_obj
			#			value = [None, None]
			if isinstance(bound, bool):
				if bound:
					value = 0, LARGE_NUMBER
				else:
					value = -LARGE_NUMBER, LARGE_NUMBER
			else:
				value = bound
				for i, v in enumerate(value):
					if i == 0 and v is None:
						value[i] = -LARGE_NUMBER
					elif i == 1 and v is None:
						value[i] = LARGE_NUMBER
			if value[0] > value[1]:
				warnings.warn('Lower bound is >= than upper bound for reaction ' + str(i) + '. Reversing...')
				value = value[::-1]
			return value

		return list(map(interpret_bound, enumerate(thermodynamic_constraints)))

	def flux_limits(self, reaction):
		self.set_objective({reaction: 1}, True)
		min_flux = self.optimize().objective_value()
		self.set_objective({reaction: 1}, False)
		max_flux = self.optimize().objective_value()

		return min_flux, max_flux

	def simplify(self, initial_objective, objective_sense, fva_gamma=1-1e-6):
		res = FluxVariabilityAnalysis(self.model).run(
			initial_objective=initial_objective, minimize_initial=objective_sense, gamma=fva_gamma)
		blocked_reaction_names = [self.reaction_names[i] for i in res.find_blocked_reactions()]
		self.remove_reactions(blocked_reaction_names)
		self.remove_orphan_metabolites()


	@property
	def gpr(self):
		return self.__gpr

	@gpr.setter
	def gpr(self, gprs, **kwargs):
		if isinstance(gprs, (list, tuple)):
			assert not (False in [isinstance(gpr, str) for gpr in gprs]), 'Non-string object found in gprs iterable'
			self.__gpr = GPRContainer(gpr_list=gprs, **kwargs)
		elif isinstance(gprs, GPRContainer):
			assert len(gprs) == len(self.bounds)
			self.__gpr = gprs
		else:
			raise(TypeError,'Can\'t use '+str(type(gprs))+' to redefine this model\'s GPRs')

	def remove_orphan_metabolites(self):
		zero_rows = [self.metabolite_names[i] for i in where((self.get_stoichiometric_matrix() == 0).all(axis=1))[0]]
		if len(zero_rows) > 0:
			self.remove_metabolites(zero_rows)

	def remove_orphan_reactions(self):
		zero_cols = [self.metabolite_names[i] for i in where((self.get_stoichiometric_matrix() == 0).all(axis=0))[0]]
		if len(zero_cols) > 0:
			self.remove_reactions(zero_cols)

	def initialize_optimizer(self):
		lb, ub = self.get_bounds_as_list()
		self.model = SteadyStateLinearSystem(self.__S, lb, ub, var_names=self.reaction_names, solver=self.solver)
		self.model.build_problem()
		self.optimizer = LinearSystemOptimizer(self.model, build=False)

	def decode_index(self, index, labels):
		if type(index) in INT_TYPES:
			return index

		elif labels is not None:
			return self.map_labels[labels][index]
		else:
			raise IndexError('Could not find specified index "' + str(index) + '".')

	def get_bounds_as_list(self):
		return list(zip(*self.bounds))

	def is_reversible_reaction(self, index):
		lb, ub = self.bounds[self.decode_index(index, 'reaction')]
		return (lb < 0 and ub > 0)

	def get_stoichiometric_matrix(self, rows=None, columns=None):
		row_index = [self.decode_index(i, 'metabolite') for i in rows] if rows else None
		col_index = [self.decode_index(i, 'reaction') for i in columns] if columns else None

		if rows and columns:
			return self.__S[row_index, col_index]
		elif rows:
			return self.__S[row_index, :]
		elif columns:
			return self.__S[:, col_index]
		else:
			return self.__S

	def set_stoichiometric_matrix(self, values, rows=None, columns=None, update_only_nonzero=False):
		row_index = [self.decode_index(i, 'metabolite') for i in rows] if rows else None
		col_index = [self.decode_index(i, 'reaction') for i in columns] if columns else None

		if self.has_context():
			self.context_managers[-1].queue_command(
				self.set_stoichiometric_matrix,
				{
					'values': self.get_stoichiometric_matrix(rows, columns),
					'rows': rows,
					'columns': columns
				}
			)

		if rows and columns:
			self.__S[row_index, col_index] = values
		elif rows:
			self.__S[row_index, :] = values
		elif columns:
			self.__S[:, col_index] = values
		else:
			self.__S = values

		row_index = row_index if row_index is not None else range(self.__S.shape[0])
		col_index = col_index if col_index is not None else range(self.__S.shape[1])

		if self.model:
			constraints = [self.model.model.constraints[i] for i in
						   row_index] if row_index else self.model.model.constraints
			vars = [self.model.model.variables[i] for i in col_index] if row_index else self.model.model.variables

			self.model.populate_constraints_from_matrix(values, constraints=constraints,
			                                            vars=vars, only_nonzero=update_only_nonzero)

	def add_metabolites(self, args, names=None):
		assert sum([n in self.reaction_names for n in names]) == 0, 'Duplicate metabolite name found!'
		if isinstance(args, list):
			rows = zeros([len(args), self.__S.shape[1]])
			for row_i in range(len(args)):
				for k, v in args[row_i].items():
					rows[row_i,self.decode_index(k, 'reaction')] = v
		elif isinstance(args, ndarray):
			if args.shape[1] == len(self.reaction_names):
				rows = args
			else:
				raise Exception('Numpy argument dimensions should be (', len(self.metabolite_names), ',). Got',
								args.shape, 'instead')
		else:
			raise ValueError('Invalid argument type: ', type(args), '. Please supply an ndarray or dict instead.')

		if self.has_context():
			self.context_managers[-1].queue_command(self.remove_metabolites, {'index': self.__S.shape[0]})

		self.__S = vstack([self.__S, rows])
		self.metabolite_names.extend(names)

		self.__update_decoder_map()

		if self.model:
			self.model.add_rows_to_model(rows.reshape([1, self.__S.shape[1]])
			                             if rows.size == self.__S.shape[1] else rows,
										 b_lb=array([0]*rows.shape[0]), b_ub=array([0]*rows.shape[0]),
										 only_nonzero=True)

	def add_reactions(self, args, bounds, names=None, gpr=None):
		assert sum([n in self.reaction_names for n in names]) == 0, 'Duplicate reaction name found!'
		if isinstance(args, (list,tuple)):
			cols = zeros([self.__S.shape[0], len(args)])
			for col_i in range(len(args)):
				for k, v in args[col_i].items():
					cols[self.decode_index(k, 'metabolite'), col_i] = v
		elif isinstance(args, ndarray):
			if args.shape[0] == len(self.metabolite_names):
				cols = args
			else:
				raise Exception('Numpy argument dimensions should be (','m',',', len(self.reaction_names),'). Got',
								str(args.shape), 'instead')
		else:
			raise ValueError('Invalid argument type: ', type(args), '. Please supply an ndarray or dict instead.')

		if self.has_context():
			self.context_managers[-1].queue_command(self.remove_reactions,
													{'index': [self.__S.shape[1]+i for i in range(cols.shape[1])]})

		self.__S = hstack([self.__S, cols])

		self.reaction_names.extend(names)
		self.bounds.extend(bounds)

		self.__update_decoder_map()
		if gpr is None:
			gprs = ['']*cols.shape[1]
		else:
			gprs = gpr
		self.gpr.add_gprs(gprs)

		lbs, ubs = zip(*bounds)
		if self.model:
			self.model.add_columns_to_model(cols, names, lbs, ubs, VAR_CONTINUOUS, only_nonzero=True)

	def add_metabolite(self, arg, name=None):
		assert not name in self.metabolite_names, 'Duplicate metabolite name found!'
		if isinstance(arg, dict):
			row = zeros(1, self.__S.shape[1])
			for k, v in arg.items():
				row[self.decode_index(k, 'reaction')] = v
		elif isinstance(arg, ndarray):
			if len(arg) == len(self.reaction_names):
				row = arg
			else:
				raise Exception('Numpy argument dimensions should be (', len(self.metabolite_names), ',). Got',
								arg.shape, 'instead')
		else:
			raise ValueError('Invalid argument type: ', type(arg), '. Please supply an ndarray or dict instead.')

		if self.has_context():
			self.context_managers[-1].queue_command(self.remove_metabolites, {'index':self.__S.shape[0]})

		self.__S = vstack([self.__S, row])
		self.metabolite_names.append(name)

		self.__update_decoder_map()

		if self.model:
			self.model.add_rows_to_model(row.reshape([1, self.__S.shape[1]]), b_lb=array([0]), b_ub=array([0]),
			                             only_nonzero=True)

	def add_reaction(self, arg, bounds, name=None, gpr=''):
		assert not name in self.reaction_names, 'Duplicate reaction name found!'
		if isinstance(arg, dict):
			col = zeros([self.__S.shape[0], 1])
			for k, v in arg.items():
				col[self.decode_index(k, 'metabolite'), 0] = v
		elif isinstance(arg, ndarray):
			if len(arg) == len(self.metabolite_names):
				col = arg.reshape(self.__S.shape[0], 1)
			else:
				raise Exception('Numpy argument dimensions should be (', len(self.reaction_names), ',). Got', arg.shape,
								'instead')
		else:
			raise ValueError('Invalid argument type: ', type(arg), '. Please supply an ndarray or dict instead.')

		if self.has_context():
			self.context_managers[-1].queue_command(self.remove_reactions,{'index':self.__S.shape[1]})

		self.__S = hstack([self.__S, col])

		self.reaction_names.append(name)
		self.bounds.append(bounds)

		self.__update_decoder_map()
		self.gpr.add_gprs([gpr])

		if self.model:
			self.model.add_columns_to_model(col, [name], [bounds[0]], [bounds[1]], VAR_CONTINUOUS, only_nonzero=True)

	def remove_reaction(self, index):
		warnings.warn('''remove_reaction will be renamed to remove_reactions in a future version to represent the
					  method\'s true behaviour''',DeprecationWarning)
		self.remove_reactions(index)

	def remove_metabolite(self, index):
		warnings.warn('''remove_metabolite will be renamed to remove_metabolites in a future version to represent the
					  method\'s true behaviour''',DeprecationWarning)
		self.remove_metabolites(index)

	def get_boundary_reactions(self, epsilon=1e-9):
		nzcoefs = where(abs(self.get_stoichiometric_matrix()) > epsilon)
		boundary_rx_ids = [k for k,v in Counter(nzcoefs[1]).items() if v == 1]
		x = array(nzcoefs)
		boundaries = {}
		for k,v in zip(*x[:,isin(x[1], boundary_rx_ids)]):
			ki, vi = self.metabolite_names[k], self.reaction_names[v]
			if ki not in boundaries:
				boundaries[ki] = [vi]
			else:
				boundaries[ki].append(vi)
		return boundaries

	def add_boundary_reactions(self, metabolites, lbs, ubs, prefix='EX_', epsilon=1e-9):
		lbs, ubs = (to_list_if_single(x, len(metabolites)) for x in [lbs, ubs])

		return self.add_reactions(args=[{k:-1} for k in metabolites],
		                   bounds=[(l,u) for l,u in zip(lbs, ubs)],
		                   names=[prefix+m for m in metabolites],
		                   gpr=['' for m in metabolites])

	def remove_reactions(self, index):
		if isinstance(index, (int, str)):
			index = [index]

		j = [self.decode_index(i, 'reaction') for i in index]

		if self.has_context():
			for decoded_index in j:
				self.context_managers[-1].queue_command(
					self.add_reaction,
					{
						'arg':self.get_stoichiometric_matrix(rows=None, columns=[decoded_index]),
						'name':self.reaction_names[decoded_index],
						'bounds':self.get_reaction_bounds(decoded_index),
						'gpr':self.gpr[decoded_index]
					}
				)


		if self.reaction_names is not None:
			new_rnames = [self.reaction_names[k] for k in range(len(self.reaction_names)) if k not in j]
			self.reaction_names = new_rnames

		new_bounds = [k for i,k in enumerate(self.bounds) if i not in j]
		self.bounds = new_bounds
		self.__S = delete(self.__S, j, axis=1)
		self.__update_decoder_map()

		self.gpr.remove_gprs(index)

		if self.model:
			self.model.remove_from_model(j, 'var')

	def remove_metabolites(self, index):
		if isinstance(index, (int, str)):
			index = [index]

		i = [self.decode_index(ii, 'metabolite') for ii in index]

		if self.has_context():
			for decoded_index in i:
				self.context_managers[-1].queue_command(
					self.add_metabolite,
					{
						'arg':self.get_stoichiometric_matrix(rows=[decoded_index], columns=None).ravel(),
						'name':self.metabolite_names[decoded_index],
					}
				)


		if self.reaction_names is not None:
			new_mnames = [k for i1,k in enumerate(self.metabolite_names) if i1 not in i]
			self.metabolite_names = new_mnames

		self.__S = delete(self.__S, i, axis=0)
		self.__update_decoder_map()

		if self.model:
			self.model.remove_from_model(i, 'const')

	def make_irreversible(self):
		lb, ub = self.get_bounds_as_list()
		Sn, nlb, nub, rx_mapping = make_irreversible_model(self.__S, lb, ub)
		rev = array([int(v[0]) for k, v in rx_mapping.items() if isinstance(v, (list, tuple))])

		gprs_irrev = ['']*len(nlb)
		for i,tup in rx_mapping.items():
			if isinstance(tup, (tuple, list)):
				gprs_irrev[tup[0]] = gprs_irrev[tup[1]] = self.gpr[i]
			else:
				gprs_irrev[tup] = self.gpr[i]

		if self.reaction_names != None:
			if len(rev) > 0:
				revnames = ['_'.join([name, BACKPREFIX]) for name in (array(self.reaction_names)[rev]).tolist()]
			else:
				revnames = []
			rnames = self.reaction_names + revnames

			model = ConstraintBasedModel(Sn, list(zip(nlb, nub)), rnames, self.metabolite_names,
										 optimizer=self.model is not None, solver=self.solver, gprs=gprs_irrev)
		else:
			model = ConstraintBasedModel(Sn, list(zip(nlb, nub)), optimizer=self.model is not None, solver=self.solver,
										 gprs=gprs_irrev)
		return model, rx_mapping

	def get_reaction_bounds(self, index):
		return self.bounds[self.decode_index(index, 'reaction')]

	def set_reaction_bounds(self, index, **kwargs):
		true_idx = self.decode_index(index, 'reaction')
		lb, ub = self.get_reaction_bounds(true_idx)

		if self.has_context():
			self.context_managers[-1].queue_command(
				self.set_reaction_bounds,
				{
					'index':true_idx,
					'lb': lb,
					'ub': ub
				}
			)

		if 'lb' in kwargs:
			lb = kwargs['lb']
		if 'ub' in kwargs:
			ub = kwargs['ub']
		bound = (lb, ub)
		self.bounds[true_idx] = bound

		if 'temporary' in kwargs and kwargs['temporary'] == False:
			self.original_bounds[self.decode_index(index, 'reaction')] = self.get_reaction_bounds(index)

		if self.model:
			var = self.model.model.variables[true_idx]
			self.model.set_variable_bounds([var], [lb], [ub])

	def set_objective(self, coef_dict, minimize=False):
		if self.model:
			if isinstance(coef_dict, dict):
				f = zeros(self.__S.shape[1], )
				self.c = f
				for k, v in coef_dict.items():
					self.c[self.decode_index(k, 'reaction')] = v
				self.model.set_objective(self.c, minimize)
			elif isinstance(coef_dict, ndarray):
				self.model.set_objective(coef_dict, minimize)
			else:
				raise TypeError('`coef_dict` must either be a dict or an ndarray')
		else:
			raise Exception('Cannot set an objective on a model without the optimizer flag as True.')

	def add_objective_constraint(self, objective_coefficients, minimize_objective, constrained_objective_perc,
	                             sink_name, sink_reaction_prefix='EX_', constrain=True, max_default_flux=1e4):
		if constrain:
			sol_prime = self.optimize(objective_coefficients, minimize=minimize_objective)
			obj_value = sol_prime.objective_value()
		else:
			obj_value = max_default_flux

		if len(objective_coefficients) > 1:
			self.add_metabolites([objective_coefficients], names=[sink_name])
			self.add_boundary_reactions([sink_name],
			                       lbs=[(constrained_objective_perc[0] * obj_value) if constrain else 0],
			                       ubs=[(constrained_objective_perc[1] * obj_value) if constrain else max_default_flux],
			                       prefix=sink_reaction_prefix)
		else:
			rx = list(objective_coefficients.keys())[0]
			if constrain:
				self.set_reaction_bounds(rx, lb=obj_value * constrained_objective_perc[0],
				                         ub=obj_value * constrained_objective_perc[1])

		return (sink_name, sink_reaction_prefix+sink_name)

	def optimize(self, coef_dict=None, minimize=False):
		cur_obj = self.model.model.objective
		if coef_dict != None:
			self.set_objective(coef_dict, minimize)
		sol = self.optimizer.optimize()

		if coef_dict != None:
			self.model.model.objective = cur_obj

		return sol

	def revert_to_original_bounds(self):
		for rx, bounds in zip(self.reaction_names, self.original_bounds):
			clb, cub = self.get_reaction_bounds(rx)
			lb, ub = bounds

			if clb != lb or cub != ub:
				self.set_reaction_bounds(rx, lb=lb, ub=ub)

	def summarize_solution(self, sol, drains, eps=1e-9):
		active_drains = {k:v for k,v in sol.var_values().items() if abs(v) > eps and k in drains}
		table = [[k]+[r for r,v in active_drains.items() if f(v)] for k,f in [('produced',lambda x: x < -eps), ('consumed',lambda x: x > eps)]]
		pretty_table_print(array(table).T.tolist(), has_header=True, header_sep=2)




import warnings
import ast
from _ast import Name

from itertools import chain
from numpy import zeros
from boolean.boolean import BooleanAlgebra
import re

GPR_GENE_RE = re.compile("\\b(?!and\\b|or\\b|AND\\b|OR\\b)[([._A-z0-9]*")

GPR_DOT = '__COBAMPGPRDOT__'
GPR_UNDERSCORE = '__COBAMPGPRUNDERSCORE__'

def symbol_encode(match_str): return '_' + match_str.replace('.', GPR_DOT) if match_str != '' else ''
def symbol_decode(match_str): return match_str[1:].replace(GPR_DOT, '.') if match_str != '' else ''


def logical_or(x):
	return sum(x) >= 1

def logical_and(x):
	return sum(x) == len(x)

def aux_apply(fx, it):
	args = [k for k in it if k is not None]
	return fx(args) if args else None

def normalize_boolean_expression(rule, simplify=False):
	BOOLEAN_ALGEBRA = BooleanAlgebra()
	try:
		expr = BOOLEAN_ALGEBRA.parse(rule).literalize().simplify()
		# Simplify first otherwise _rdistributive() may take forever.
		operation_example = BOOLEAN_ALGEBRA.OR(BOOLEAN_ALGEBRA.TRUE, BOOLEAN_ALGEBRA.FALSE)
		expr = BOOLEAN_ALGEBRA._rdistributive(expr, operation_example)
		if simplify:
			expr = expr.simplify()
		#bool_expression = BOOLEAN_ALGEBRA.normalize(expression, BOOLEAN_ALGEBRA.OR)
		return str(expr).replace('&', ' and ').replace('|', ' or ')
	except Exception as e:
		warnings.warn('Could not normalize this rule: ' + rule)
		return rule

def convert_gpr_to_list(gpr, apply_fx=str, or_char='or', and_char='and'):
	proteins = list(
		map(
			lambda x: x.strip().replace('(', '').replace(')', ''),
			apply_fx(gpr).split(or_char)
		)
	)
	rules = [[s.strip() for s in x.split(and_char) if s.strip() != ''] for x in proteins]
	return rules


class GPRContainer(object):
	def __init__(self, gpr_list, apply_fx=str, or_char='or', and_char='and', ttg_ratio=20):
		self.apply_fx = apply_fx
		self.__or_char, self.__and_char = or_char, and_char
		self.ttg_ratio = ttg_ratio
		self.__initialize(gpr_list)



	def __initialize(self, gpr_list):
		self.__gprs = []
		self.__gpr_list = []
		self.add_gprs(gpr_list)
		self.__update_genes()

	def add_gprs(self, gpr_list):
		gprs = [self.__preprocess_gprs(gp, token_to_gene_ratio=self.ttg_ratio) if gp != '' else ''
				for i,gp in enumerate(gpr_list)]
		gpr_list = [convert_gpr_to_list(g, apply_fx=str, or_char=self.__or_char, and_char=self.__and_char) for g in
		 gprs]

		self.__gprs.extend(gprs)
		self.__gpr_list.extend(gpr_list)
		self.__update_genes()

	def remove_gprs(self, indices):
		self.__gprs, self.__gpr_list = zip(*[(self.__gprs[i], self.__gpr_list[i]) for i in range(len(self.__gprs))
											 if i not in indices])
		self.__gprs, self.__gpr_list = [list(k) for k in [self.__gprs, self.__gpr_list]]
		self.__update_genes()

	def __update_genes(self):
		self.__genes = tuple(set(list(chain(*chain(*self.__gpr_list)))))
		self.__gene_to_index_mapping = dict(zip(self.__genes, range(len(self.__genes))))
		self.__safe_gpr_decoder = {symbol_encode(g):g for g in self.get_genes()}

	def get_safe_gpr_decoder(self):
		return self.__safe_gpr_decoder

	def get_ast(self, index, decode=False):
		gpr = self[index]
		encoded = GPR_GENE_RE.sub(lambda x: symbol_encode(x.group()), gpr)
		if decode:
			for node in filter(lambda x: isinstance(x, Name), ast.walk(tree)):
				node.id = gpr_eval.get_safe_gpr_decoder()[node.id]
		return ast.parse(encoded).body[0]

	def __preprocess_gprs(self, gpr_str, token_to_gene_ratio=20):
		def fix_name(gpr_string):
			matches = [k for k in GPR_GENE_RE.finditer(gpr_string) if k.span()[0] - k.span()[1] != 0]
			unique_tokens = set([m.string for m in matches])
			return GPR_GENE_RE.sub(lambda x: symbol_encode(x.group()), gpr_string), len(matches), len(unique_tokens), unique_tokens
		# gpr_list = []
		# for gpr_str in gpr_string_list:
		rule, num_matches, num_unique_tokens, unique_tokens = fix_name(gpr_str)
		if self.apply_fx:
			rule = self.apply_fx(rule)
		if (num_unique_tokens > 0) and (num_matches // num_unique_tokens) < token_to_gene_ratio:
			rule = normalize_boolean_expression(rule)
		else:
			warnings.warn(
				'Will not normalize rules with more than ' + str(token_to_gene_ratio) + ' average tokens per gene')

		return GPR_GENE_RE.sub(lambda x: symbol_decode(x.group()), rule)


	def __len__(self):
		return len(self.__gpr_list)

	def __getitem__(self, item):
		return self.__gprs[item]

	def gpr_has_string(self, index, string):
		return string in self.__gprs[index]

	def get_gpr_as_lists(self, index):
		return self.__gpr_list[index]

	def or_char(self):
		return self.__or_char

	def and_char(self):
		return self.__and_char

	def get_genes(self):
		return self.__genes

	def eval_gpr(self, index, state, or_fx=logical_or, and_fx=logical_and):
		return aux_apply(or_fx,
						 [aux_apply(and_fx, [state[x] for x in gs if x in state.keys()]) for gs in
						  self.__gpr_list[index]])

	def associated_genes(self, index):
		return list(chain(*self.__gpr_list[index]))

	def associated_gene_matrix(self):
		B = zeros([len(self.__genes), len(self.__gpr_list)])
		row_ind, col_ind = [], []
		for i in range(len(self)):
			gene_indexes = [self.__gene_to_index_mapping[k] for k in self.associated_genes(i)]
			row_ind += gene_indexes
			col_ind += [i] * len(gene_indexes)
		B[row_ind, col_ind] = 1
		return B

if __name__ == '__main__':
	gprs = [
		'11.1 and G2 or G2.7',
		'G3 or G4',
		'(G1 and G2) or (G3 and G4)',
		'G5',
		'G6'
	]
	# gprs = [model.reactions[i].gene_reaction_rule for i in range(len(model.reactions))]
	gpr_eval = GPRContainer(gprs)
	for i in range(len(gprs)):
		print(gprs[i], gpr_eval.get_gpr_as_lists(i))

	tree = gpr_eval.get_ast(0)



import abc
import warnings

import numpy as np
from numpy import where

from cobamp.core.models import ConstraintBasedModel
from cobamp.gpr.core import GPRContainer
from cobamp.utilities.parallel import batch_run, cpu_count

MAX_PRECISION = 1e-10


## TODO: Add simulation and result functions on the wrappers module to automatically detect which one to use
class ConstraintBasedModelSimulator(object):
    def __init__(self, model, simulation_function, result_function):
        self.__model = model
        self.__simulation_function = simulation_function
        self.__result_function = result_function

    def simulate(self, func, bound_change=None, objective_coefficient=None, minimize=None, func_args=None):
        self.__simulation_function(self.__model, func, bound_change, objective_coefficient, minimize,
                                   self.__result_function, func_args)

    def batch_simulate(self, func, bound_changes, objective_coefficients, minimize, func_args=None, mp_threads=None):
        mp_params = {'model':self.__model, 'func_args':func_args}

        def check_multiple_inputs(bound_changes, objective_coefficients, minimize):
            arg_list = bound_changes, objective_coefficients, minimize
            corrected_args = []
            max_len = max(map(len,arg_list))
            for arg in arg_list:
                if len(arg) == 1:
                    corrected_args.append([arg[0]]*max_len)
                elif len(arg) == max_len:
                    corrected_args.append(arg)
                else:
                    raise Exception('One of the arguments contains more than 1 and less than '+str(max_len)+' items.')
            return corrected_args


        def batch_simulation_function(sequence, params):
            model_mp = params['model']
            func_args_mp = params['func_args']
            bc, oc, min = sequence
            return self.__simulation_function(model_mp, func, bc, oc, min, self.__result_function, func_args_mp)

        sequence = list(zip(*check_multiple_inputs(bound_changes, objective_coefficients, minimize)))
        return batch_run(batch_simulation_function, sequence, mp_params,
                         threads=cpu_count() if mp_threads is None else mp_threads)

class AbstractObjectReader(object):
    """
    An abstract class for reading metabolic model objects from external frameworks, and extracting the data needed for
    pathway analysis methods. Also deals with name conversions.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, model, gpr_and_char='and', gpr_or_char='or', gpr_gene_parse_function=str, ttg_ratio=20):
        """
        Parameters

        ----------

            model: A Model instance from the external framework to use. Must be registered in the dict stored as
            external_wrappers.model_readers along with its reader.

        """
        self.model = model
        self.initialize(gpr_and_char, gpr_or_char, gpr_gene_parse_function, ttg_ratio)

    def initialize(self, gpr_and_char='and', gpr_or_char='or', gpr_gene_parse_function=str, ttg_ratio=20):
        """
            This method re-initializes the class attributes from the current state of self.model
        """

        self.r_ids, self.m_ids = self.get_reaction_and_metabolite_ids()
        self.rx_instances = self.get_rx_instances()
        self.S = self.get_stoichiometric_matrix()
        self.lb, self.ub = tuple(zip(*self.get_model_bounds(False)))
        self.irrev_bool = self.get_irreversibilities(False)
        self.irrev_index = self.get_irreversibilities(True)
        self.bounds_dict = self.get_model_bounds(True)
        self.gene_protein_reaction_rules = gpr_and_char, gpr_or_char, gpr_gene_parse_function, ttg_ratio
        self.__gpr_read_params = gpr_and_char, gpr_or_char, gpr_gene_parse_function, ttg_ratio
    @property
    def gene_protein_reaction_rules(self):
        return self.__gene_protein_reaction_rules


    @gene_protein_reaction_rules.setter
    def gene_protein_reaction_rules(self, value):
        and_char, or_char, apply_fx, ttg_ratio = value
        self.__gene_protein_reaction_rules = GPRContainer(
            gpr_list=self.get_model_gpr_strings(),
            and_char=and_char, or_char=or_char, apply_fx=apply_fx, ttg_ratio=ttg_ratio
        )

    @abc.abstractmethod
    def get_stoichiometric_matrix(self):
        """
        Returns a 2D numpy array with the stoichiometric matrix whose metabolite and reaction indexes match the names
        defined in the class variables r_ids and m_ids
        """
        pass

    @abc.abstractmethod
    def get_model_bounds(self, as_dict, separate_list):
        """
        Returns the lower and upper bounds for all fluxes in the model. This either comes in the form of an ordered list
        with tuples of size 2 (lb,ub) or a dictionary with the same tuples mapped by strings with reaction identifiers.

        Parameters

        ----------

            as_dict: A boolean value that controls whether the result is a dictionary mapping str to tuple of size 2
            separate: A boolean value that controls whether the result is two numpy.array(), one for lb and the other\n
            to ub
        """
        pass

    @abc.abstractmethod
    def get_irreversibilities(self, as_index):
        """
        Returns a vector representing irreversible reactions, either as a vector of booleans (each value is a flux,
        ordered in the same way as reaction identifiers) or as a vector of reaction indexes.

        Parameters

        ----------

            as_dict: A boolean value that controls whether the result is a vector of indexes

        """
        pass

    @abc.abstractmethod
    def get_rx_instances(self):
        """
        Returns the reaction instances contained in the model. Varies depending on the framework.
        """
        pass

    @abc.abstractmethod
    def get_reaction_and_metabolite_ids(self):
        """
        Returns two ordered iterables containing the metabolite and reaction ids respectively.
        """
        pass

    # @abc.abstractmethod
    # def get_model_genes(self):
    # 	"""
    #
    # 	Returns the identifiers for the genes contained in the model
    #
    # 	"""

    @property
    def genes(self):
        return self.gene_protein_reaction_rules.get_genes()


    def reaction_id_to_index(self, id):
        """
        Returns the numerical index of a reaction when given a string representing its identifier.

        Parameters

        ----------

            id: A reaction identifier as a string

        """
        return self.r_ids.index(id)

    def metabolite_id_to_index(self, id):
        """
        Returns the numerical index of a metabolite when given a string representing its identifier.

        Parameters

        ----------

            id: A metabolite identifier as a string

        """
        return self.m_ids.index(id)

    def get_gene_protein_reaction_rule(self, id):
        return self.gene_protein_reaction_rules[id]

    def convert_constraint_ids(self, tup, yield_constraint):
        if yield_constraint:
            constraint = tuple(list(map(self.reaction_id_to_index, tup[:2])) + list(tup[2:]))
        else:
            constraint = tuple([self.reaction_id_to_index(tup[0])] + list(tup[1:]))
        return constraint


    def get_reaction_scores(self, expression, and_fx=min, or_fx=max, as_vector=False):
        exp_map = {rx: self.gene_protein_reaction_rules.eval_gpr(i, expression, or_fx=or_fx, and_fx=and_fx)
                   for i,rx in enumerate(self.r_ids)}

        if as_vector:
            return [exp_map[k] for k in self.r_ids]
        else:
            return exp_map

    #@warnings.warn('g2rx will be deprecated in a future release. Use the get_reaction_scores method instead',
    #			   DeprecationWarning)
    def g2rx(self, expression, and_fx=min, or_fx=max, as_vector=False, apply_fx=str):
        warnings.warn('g2rx will be deprecated in a future release. Use the get_reaction_scores method instead',
                       DeprecationWarning)
        return self.get_reaction_scores(expression, and_fx, or_fx, as_vector)


    @abc.abstractmethod
    def get_model_gpr_strings(self):
        pass


    def to_cobamp_cbm(self, solver=None):
        and_char, or_char, gpr_gene_parse_function, ttg_ratio = self.__gpr_read_params
        ngprs = GPRContainer(
            gpr_list=self.get_model_gpr_strings(),
            and_char=and_char, or_char=or_char, apply_fx=gpr_gene_parse_function, ttg_ratio=ttg_ratio)

        return ConstraintBasedModel(
            S=self.get_stoichiometric_matrix(),
            thermodynamic_constraints=[tuple(float(k) for k in l) for l in self.get_model_bounds()],
            reaction_names=self.r_ids,
            metabolite_names=self.m_ids,
            optimizer= (solver == True) or (solver is not None and solver != False),
            solver=solver if solver not in (True, False) else None,
            gprs=ngprs
        )

class MatFormatReader(AbstractObjectReader):
    def get_stoichiometric_matrix(self):
        return (self.model['S'][0][0]).toarray()

    def get_model_bounds(self, as_dict=False, separate_list=False):
        lb, ub = [(self.model[k][0][0]).ravel() for k in ('lb', 'ub')]
        tuples = [(r, (l, u)) for r, l, u in zip(self.r_ids, lb, ub)]
        if as_dict:
            return dict(tuples)
        else:
            if separate_list:
                return lb, ub
            else:
                return tuple([(l, u) for l, u in zip(lb, ub)])

    def get_irreversibilities(self, as_index):
        if 'rev' in self.model.dtype.names:
            bv = (self.model['rev'][0][0]).ravel().astype(bool)
        else:
            bv = np.array([(l >= 0 and u >= 0) or (l <= 0 and u <= 0) for l, u in zip(self.lb, self.ub)]).astype(bool)
        if as_index:
            return where(bv)[0]
        else:
            return bv

    def get_rx_instances(self):
        pass

    def get_reaction_and_metabolite_ids(self):
        return [[k[0][0] for k in self.model[t][0][0]] for t in ['rxns', 'mets']]

    def get_model_genes(self):
        return set([k[0][0] for k in self.model['genes'][0][0]])

    def get_model_gpr_strings(self):
        return [k[0][0] if len(k[0]) > 0 else '' for k in self.model['grRules'][0][0]]


from functools import partial


class CommandHistory(object):
	def __init__(self):
		self.commands = []

	def execute_last(self):
		self.commands.pop()()

	def execute_first(self):
		self.commands.pop(0)()

	def execute_all(self, forward=False):
		while len(self.commands) > 0:
			(self.execute_first if forward else self.execute_last)()

	def queue_command(self, func, args):
		self.commands.append(partial(func, **args))


from numpy import where, zeros, array, vstack, unique, eye, hstack, ones
from cobamp.wrappers.method_wrappers import KShortestGenericMCSEnumeratorWrapper
from cobamp.core.models import ConstraintBasedModel
from functools import reduce

def filled_vector(dim, index, fill_value=1):
	fvec = zeros(dim)
	fvec[index] = fill_value
	return fvec


class GeneMatrixBuilder(object):
	def __init__(self, gpr_evaluator):
		self.__gpr_evaluator = gpr_evaluator
		self.__B, self.__b = self.get_association_matrix()

	def get_association_matrix(self):
		B = self.__gpr_evaluator.associated_gene_matrix()
		b = B.sum(axis=0)
		return B, b

	@staticmethod
	def gpr_single_gene(gpr_list):
		return len(gpr_list) == 1 and len(gpr_list[0]) == 1

	@staticmethod
	def gpr_only_and_rules(gpr_list):
		return len(gpr_list) == 1 and len(gpr_list[0]) > 1

	@staticmethod
	def gpr_only_or_rules(gpr_list):
		return len(gpr_list) > 1 and not (False in ([len(g) <= 1 for g in gpr_list]))

	def optimize_dual_gpr_graph(self, F):
		# S -> H -> subset of G
		cbm = self.get_gpr_model(F)
		dual_mat = zeros([F.shape[1],len(cbm.reaction_names)])
		dual_mat[:F.shape[1],:F.shape[1]] = eye(F.shape[1])
		wrp = KShortestGenericMCSEnumeratorWrapper(
			model=cbm, target_flux_space_dict={'GPRs':(1,None)}, target_yield_space_dict={},
			dual_matrix=dual_mat, dual_var_mapper={i:i for i in range(F.shape[1])}, stop_criteria=F.shape[1],
			algorithm_type='kse_populate'
		)
		enum = wrp.get_enumerator()
		sols = list(map(lambda d: array(list(d.keys())),reduce(lambda x,y: x+y, enum)))
		return sols

	@staticmethod
	def get_gpr_model(F):
		c,g = F.shape

		S = vstack([hstack(r) for r in [
			[eye(g), -F.T, zeros([g,c+1])],
			[zeros([c, g]), eye(c), -eye(c), zeros([c, 1])],
			[zeros([1, c+g]), ones([1,c]), -ones([1,1])]
		]])

		bounds = [[0, None] for _ in range(S.shape[1])]
		mn = ['MG'+str(i) for i in range(g)] + ['MC'+str(i) for i in range(c)] + ['GPR']
		rn = ['SG'+str(i) for i in range(g)] + ['SC'+str(i) for i in range(c)] + ['OC'+str(i) for i in range(c)] + \
			 ['GPRs']
		return ConstraintBasedModel(S, bounds, reaction_names=rn, metabolite_names=mn)

	def get_GF_matrices(self):
		gpr_eval = self.__gpr_evaluator
		genes = dict(zip(gpr_eval.get_genes(), range(len(gpr_eval.get_genes()))))
		total_g, total_f = [], []
		for i in range(len(gpr_eval) - 1):
			gpr_as_list = gpr_eval.get_gpr_as_lists(i)
			# AND means new row
			# OR means all active indices in the same row
			if not (len(gpr_as_list) == 0 or (len(gpr_as_list) == 1 and len(gpr_as_list[0]) == 0)):
				alt_case = [f(gpr_as_list) for f in
							[self.gpr_single_gene, self.gpr_only_and_rules, self.gpr_only_or_rules]]
				cur_g, cur_f = [], []
				if not (True in alt_case) and len(gpr_as_list) > 0:
					for complex in gpr_as_list:
						if len(complex) > 0:
							flist = filled_vector(len(genes), [genes[g] for g in complex], fill_value=1)
							glist = filled_vector(len(gpr_eval), [i], fill_value=1)
							cur_g.append(glist)
							cur_f.append(flist)

					F_sub_full = vstack(cur_f)
					F_genes = unique(F_sub_full.nonzero()[1])
					F_temp = F_sub_full[:, F_genes]
					row_inds = [F_genes[k] for k in self.optimize_dual_gpr_graph(F_temp)]
					cur_f = [filled_vector(len(genes), inds, fill_value=1) for inds in row_inds]
					cur_g = [filled_vector(len(gpr_eval), [i], fill_value=1) for _ in row_inds]
				else:
					if len(gpr_as_list) > 0  and len(gpr_as_list[0]) > 0:
						if alt_case[2]:
							flist = [filled_vector(len(genes), [genes[c[0]] for c in gpr_as_list], fill_value=1)]
							glist = [filled_vector(len(gpr_eval), [i], fill_value=1)]
						elif alt_case[1]:
							flist = [filled_vector(len(genes), [genes[g]], fill_value=1) for g in gpr_as_list[0]]
							glist = [filled_vector(len(gpr_eval), [i], fill_value=1)]*len(gpr_as_list[0])
						elif alt_case[0]:
							flist = [filled_vector(len(genes), [genes[gpr_as_list[0][0]]], fill_value=1)]
							glist = [filled_vector(len(gpr_eval), [i], fill_value=1)]
						else:
							flist, glist = [], []
						cur_g.extend(glist)
						cur_f.extend(flist)
				total_f.extend(cur_f)
				total_g.extend(cur_g)

		G, F = [vstack(l).astype(bool) for l in [total_g, total_f]]

		Fm, old_indices, reverse = unique(F, axis=0, return_index=True, return_inverse=True)
		Frev_dict = {k: where(reverse == k)[0].tolist() for k in unique(reverse)}
		Gm = vstack([G[idx, :].any(axis=0) for k, idx in Frev_dict.items()])
		rev_gmap = {v: k for k, v in genes.items()}
		irreducible_gene_map = {' and '.join(rev_gmap[k] for k in row.nonzero()[0]): i
								for i, row in enumerate(Fm)}

		gene_sets = [set(f.nonzero()[0]) for f in Fm]
		weights = [len(g) for g in gene_sets]
		F_deps = {i: {j for j, h in enumerate(gene_sets) if len(h & g) == len(g) and i != j} for i, g in
				  enumerate(gene_sets)}


		return [x.astype(int) for x in [Gm, Fm]] + [irreducible_gene_map, F_deps, weights]



if __name__ == '__main__':
	from cobamp.gpr.core import GPRContainer, logical_and, logical_or

	gprs = [
		'g1',
		'g2',
		'g2',
		'g3 and g4',
		'g2 and g5',
		'g3 or g6',
		'(g2 and (g5 or g6)) or g7',
		''
	]
	gpr_eval = GPRContainer(gprs)
	gmb = GeneMatrixBuilder(gpr_eval)
	G, F, gene_mapping = gmb.get_GF_matrices()
	ord_map = [gene_mapping['g'+str(i)] for i in range(1,8)]
	G[:,ord_map]
	F[:,ord_map]
	# revmap = {v:k for k,v in gene_mapping.items()}
	# [[revmap[k] for k in r.nonzero()[0]] for r in F]






from importlib import import_module
from .core import MatFormatReader, ConstraintBasedModelSimulator
from .cobamp import CobampModelObjectReader

model_readers = {'numpy': MatFormatReader, 'cobamp.core.models': CobampModelObjectReader}

external_frameworks = {'cobra':'COBRAModelObjectReader', 'reframed':'FramedModelObjectReader'}
external_framework_readers = {}

available_readers_dict = {
	'numpy':'MatFormatReader',
	'cobamp.core.models':'CobampModelObjectReader',
	'cobra.core.model': 'COBRAModelObjectReader',
	'framed.model.cbmodel': 'FramedModelObjectReader'
}
for module_name, reader_name in external_frameworks.items():
	try:
		module = import_module(module_name, '')
# 		print(reader_name,'is available for',module_name)
		cobamp_module = import_module('.'+module_name, package='cobamp.wrappers')
		reader_class = getattr(cobamp_module, reader_name)
		globals().update({module_name: cobamp_module, reader_name: reader_class})
		external_framework_readers[reader_name] = reader_class
	except Exception as e:
# 		print(reader_name,'could not be loaded for',module_name)
		reader = f'{reader_name} could not be loaded for {module_name}'		


available_readers_dict = {k:v for k,v in available_readers_dict.items() if v in globals()}

def get_model_reader(model_obj, **kwargs):
	if type(model_obj).__module__ in available_readers_dict.keys():
		return globals()[available_readers_dict[type(model_obj).__module__]](model_obj, **kwargs)
	else:
		raise TypeError('model_obj has an unknown type that could not be read with cobamp:', type(model_obj).__module__)


from . import method_wrappers
from .method_wrappers import KShortestEFMEnumeratorWrapper, KShortestEFPEnumeratorWrapper, KShortestMCSEnumeratorWrapper
from .method_wrappers import KShortestGenericMCSEnumeratorWrapper, KShortestGeneticMCSEnumeratorWrapper


